"""
See:
https://stackoverflow.com/questions/17223301/python-multiprocessing-is-it-possible-to-have-a-pool-inside-of-a-pool
https://stackoverflow.com/questions/52948447/error-group-argument-must-be-none-for-now-in-multiprocessing-pool
https://github.com/nipy/nipype/pull/2754
"""

import multiprocessing.pool


class Pool(multiprocessing.pool.Pool):
    # noinspection PyPep8Naming
    def Process(self, *args, **kwargs):
        proc = super(Pool, self).Process(*args, **kwargs)

        class NonDaemonProcess(proc.__class__):
            """Monkey-patch process to ensure it is never daemonized"""

            @property
            def daemon(self):
                return False

            @daemon.setter
            def daemon(self, val):
                pass

        proc.__class__ = NonDaemonProcess
        return proc



# noinspection PyPep8
"""
    Convert MCNP meshtal file to a number of npz files, one for each meshtal.

    Usage:

        mesh2npz ] [-s TALLIES] [-p PREFIX] MESH_TALLY...
        mesh2npz --version | -h | --help

    Options:
        -h, --help  print this message and exit
        --version print the version and exit
        -s, --select TALLIES            - comma separated list of tallies to extract from the mesh file
        -p, --prefix PREFIX             - prefix to prepend output files (default: "npz/"),
                                          output files are also prepended with MESH_TALLY file base name
        --override                      - override existing output files, default(false)
        --update                        - override existing output files, which are older than the source file

    Arguments:
        MESH_TALLY... - files to load mesh tallies from (default: all the .m files in current folder)


    Features:
        Fails if, an output file exist and neither --override, nor --update options is specified in command line
        Uses standard mckit_meshes module logging (see logging_cfg docs).
        Default log file mesh_2_npz.log.
"""
from __future__ import annotations

import typing as t

import logging

from pathlib import Path

import mckit_meshes.fmesh as fmesh

from ...utils.io import check_if_path_exists

__LOG = logging.getLogger(__name__)


def revise_mesh_tallies(mesh_tallies) -> t.List[Path]:
    if mesh_tallies:
        return list(map(Path, mesh_tallies))

    cwd = Path.cwd()
    rv = list(cwd.glob("*.m"))
    if not rv:
        errmsg = "No .m-files found in directory '{}', nothing to do.".format(
            cwd.absolute()
        )
        __LOG.warning(errmsg)
    return rv


def mesh2npz(
    prefix: str | Path, mesh_tallies: t.Iterable[str | Path], override: bool = False
) -> None:
    """Convert MCNP meshtal file to a number of npz files, one for each mesh tally."""
    mesh_tallies = revise_mesh_tallies(mesh_tallies)
    single_input = len(mesh_tallies) == 1
    prefix = Path(prefix)
    for m in mesh_tallies:
        m = Path(m)
        if single_input:
            p = prefix
        else:
            p = prefix / m.stem
        __LOG.info("Processing {}".format(m))
        __LOG.debug("Saving tallies with prefix {}".format(prefix))
        p.mkdir(parents=True, exist_ok=True)
        with m.open() as stream:
            fmesh.m_2_npz(
                stream,
                prefix=p,
                check_existing_file_strategy=check_if_path_exists(override),
            )



"""Apply function to cartesian product of arrays."""
from typing import Any, Callable, Dict, Sized

from itertools import product

import numpy as np


def cartesian_product(
    *arrays: Sized, aggregator: Callable[[Any, Dict], Any], **kw: Dict[Any, Any]
) -> np.ndarray:
    """Computes transformations of cartesian product of all the elements in arrays.

    Args:
        arrays:  The arrays to product.
        aggregator: Callable to handle an item from product iterator.
            The first parameter of the callable is tuple of current product item.
            May return scalar or numpy ndarray.
        kw: keyword arguments to pass to aggregator

    Returns:
        ret: Numpy array with dimension of arrays and one more
            additional dimensions for their cartesian product.
    """
    res = np.stack([aggregator(x, **kw) for x in product(*arrays)])
    shape = tuple(map(len, arrays))
    if 1 < len(res.shape):  # the aggregation result is vector
        shape = shape + res.shape[1:]
    return res.reshape(shape)



"""Convert npz files to VTK vtr-files."""

from __future__ import annotations

import typing as t

import logging

from pathlib import Path

import mckit_meshes.fmesh as fmesh

from ...utils.io import check_if_path_exists

__LOG = logging.getLogger(__name__)


def revise_npz_files(npz_files: t.Optional[t.Iterable[t.Any]]) -> t.List[Path]:
    """Use specified list of file to process, if any, or find them in current directory.

    Args:
        npz_files: npz files to process passed from command line, optional.

    Returns:
        List of the specified or found files as Path objects.
    """
    if npz_files:
        return list(map(Path, npz_files))

    cwd = Path.cwd()
    rv = list(cwd.glob("*.npz"))
    if not rv:
        errmsg = f"No .npz-files found in directory '{cwd.absolute()}', nothing to do."
        __LOG.warning(errmsg)
    return rv


def npz2vtk(
    prefix: str | Path, npz_files: t.Iterable[str | Path], override: bool = False
) -> None:
    """Convert MCNP meshtal file to a number of npz files, one for each mesh tally.

    Args:
        prefix: output directory
        npz_files: files to process, optional
        override: define behaviour when output file, exists, default - rise FileExistsError.
    """
    npz_files = revise_npz_files(npz_files)
    prefix = Path(prefix)
    file_exists_strategy = check_if_path_exists(override)
    for npz in npz_files:
        npz = Path(npz)
        __LOG.info("Processing {}".format(npz))
        __LOG.debug("Saving VTK file with prefix {}".format(prefix))
        prefix.mkdir(parents=True, exist_ok=True)
        mesh = fmesh.FMesh.load_npz(npz)
        vtk_file_stem = f"{prefix / str(mesh.name)}"
        vtk_file_name = (
            vtk_file_stem + ".vtr"
        )  # TODO dvp: revise this when it comes to saving structured mesh
        file_exists_strategy(vtk_file_name)
        vtk = mesh.save2vtk(vtk_file_stem)
        __LOG.info("Saved VTK to {}", vtk)



"""Common mesh geometry specification classes and functions.

## Relative or absolute coordinates

    There are variations when coordinates are presented as relative to origin
    or absolute. This depends on is output for specification, or input/output
    to Weight of Meshtal files and is it cartesian or cylinder mesh.

    Cartesian:

    |       |  wwinp   | meshtal  |
    | ===== | =======  | ======== |
    |  spec | relative | absolute (but origin is extracted to separate item) |
    | ----- | -------  | -------- |
    |  file | relative | absolute |

    Cylinder:

    |       |  wwinp   | meshtal  |
    | ===== | =======  | ======== |
    |  spec | relative | relative |
    | ----- | -------  | -------- |
    |  file | relative | relative |

    The new callers are to use local_coordinates converter to avoid difficulties.
    For the old callers we will use ZERO_ORIGIN for Geometry Specification being
    used in FMesh.

"""
from __future__ import annotations

from typing import Final, Iterable, List, Optional, Sequence, TextIO, Tuple, Union, cast

import abc

from dataclasses import dataclass, field

import mckit_meshes.utils as ut
import numpy as np
import numpy.linalg as linalg

from mckit_meshes.utils.cartesian_product import cartesian_product

_2PI: Final[float] = 2.0 * np.pi
_1_TO_2PI: Final[float] = 1 / _2PI
__DEG_2_RAD: Final[float] = np.pi / 180.0
CARTESIAN_BASIS: Final[np.ndarray] = np.eye(3, dtype=np.double)
(NX, NY, NZ) = CARTESIAN_BASIS

DEFAULT_AXIS: Final[np.ndarray] = NZ
DEFAULT_VEC: Final[np.ndarray] = NX


ZERO_ORIGIN: Final[np.ndarray] = np.zeros((3,), dtype=np.double)


def as_float_array(array) -> np.ndarray:
    """Convert any sequence of numbers to numpy array of floats.

    Args:
        array: Anything that can be converted to numpy ndarray.

    Returns:
        np.ndarray:  either original or conversion.

    """
    return np.asarray(array, dtype=float)


@dataclass(eq=False)
class AbstractGeometrySpecData:
    """Data mixin for :py:class:`AbstractGeometrySpec`.

    Provides reusable data fields.

    """

    ibins: np.ndarray
    jbins: np.ndarray
    kbins: np.ndarray
    origin: np.ndarray = field(default_factory=lambda: ZERO_ORIGIN.copy())

    def __post_init__(self) -> None:
        """Force a caller provided data as numpy arrays.

        Raises:
            TypeError: if any of the fields is not a numpy array.
        """
        for b in self.bins:
            if not isinstance(b, np.ndarray):
                raise TypeError(f"Expected numpy array, actual {b[0]}...{b[-1]}")

    def __hash__(self) -> int:
        return hash(self.bins)

    def __eq__(self, other: "AbstractGeometrySpecData") -> bool:
        if not isinstance(other, AbstractGeometrySpecData):
            return False
        a, b = self.bins, other.bins
        return len(a) == len(b) and arrays_equal(zip(a, b))

    @property
    def bins(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Pack the fields to tuple.

        Returns:
            tuple of origin and bins.
        """
        return self.origin, self.ibins, self.jbins, self.kbins


class AbstractGeometrySpec(AbstractGeometrySpecData, abc.ABC):
    """Common base for rectilinear and cylinder mesh specifications."""

    @property
    @abc.abstractmethod
    def cylinder(self) -> bool:
        """Is this an instance of a cylinder mesh specification?"""
        ...

    @abc.abstractmethod
    def local_coordinates(self, points: np.ndarray) -> np.ndarray:
        """Convert points coordinates to local system.

        Args:
            points: ... with global coordinates
        """
        ...

    @abc.abstractmethod
    def get_mean_square_distance_weights(self, point) -> float:
        """Estimate weights as a voxel mean square distance from the point.

        Args:
            point: ... from where to compute distance
        """
        ...

    @abc.abstractmethod
    def calc_cell_centers(self) -> np.ndarray:
        """Calculate cell (voxel) centers."""
        ...

    @abc.abstractmethod
    def print_geom(self, io: TextIO, indent: str) -> None:
        """Print geometry specification.

        Args:
            io: stream to print to
            indent: indent to insert before lines

        """
        ...

    # Generic methods

    @property
    def bins_shape(self) -> Tuple[int, int, int]:
        """Shape of data corresponding to spatial bins.

        Returns:
            Tuple with the data shape.
        """
        return (self.ibins.size - 1), (self.jbins.size - 1), (self.kbins.size - 1)

    @property
    def bins_size(self) -> int:
        """Size of data corresponding to spatial bins.

        Returns:
            int: number of voxels
        """
        return (self.ibins.size - 1) * (self.jbins.size - 1) * (self.kbins.size - 1)

    @property
    def boundaries(self) -> np.ndarray:
        """Corners or min, max values of bins."""
        return np.vstack(
            (
                self.ibins[[0, -1]],
                self.jbins[[0, -1]],
                self.kbins[[0, -1]],
            ),
        )

    @property
    def boundaries_shape(self) -> Tuple[int, int, int]:
        """Bins (boundaries) shape."""
        return self.ibins.size, self.jbins.size, self.kbins.size

    def surrounds_point(self, x: float, y: float, z: float, local: bool = True) -> bool:
        """Check if the point (x,y,z) is within the volume of mesh.

        By default, assumes that the point is given in local coordinates.
        """
        if not local:
            x, y, z = self.local_coordinates(np.array([x, y, z], dtype=float))
        (xmin, xmax), (ymin, ymax), (zmin, zmax) = self.boundaries
        return cast(bool, (xmin < x < xmax) and (ymin < y < ymax) and (zmin < z < zmax))

    def select_indexes(
        self, *, i_values=None, j_values=None, k_values=None
    ) -> Tuple[
        Union[int, slice, np.ndarray],
        Union[int, slice, np.ndarray],
        Union[int, slice, np.ndarray],
    ]:
        """Select indices for data corresponding to given spatial values

        Args:
            i_values: indices along i (X or R) dimension
            j_values: ... along j (Y or Z)
            k_values: ... along k (Z or Theta)

        Returns:
            see :py:func:`select_indexes()`
        """
        return (
            select_indexes(self.ibins, i_values),
            select_indexes(self.jbins, j_values),
            select_indexes(self.kbins, k_values),
        )

    def print(self, io: TextIO, columns: int = 6) -> None:
        indent = " " * 8
        self.print_geom(io, indent)
        print(indent, "origin=", " ".join(format_floats(self.origin)), sep="", file=io)
        _print_bins(indent, "i", self.ibins, io, columns=columns)
        _print_bins(indent, "j", self.jbins, io, columns=columns)
        _print_bins(indent, "k", self.kbins, io, columns=columns)


class CartesianGeometrySpec(AbstractGeometrySpec):

    # TODO dvp: add transformation

    @property
    def cylinder(self) -> bool:
        return False

    @property
    def x(self) -> np.ndarray:
        return self.ibins

    @property
    def y(self) -> np.ndarray:
        return self.jbins

    @property
    def z(self) -> np.ndarray:
        return self.kbins

    def local_coordinates(self, points: np.ndarray) -> np.ndarray:
        assert points.shape[-1] == 3, "Expected cartesian point array or single point"
        if self.origin is not ZERO_ORIGIN:
            return cast(
                np.ndarray, points - ZERO_ORIGIN
            )  # TODO dvp: recall what is this subtraction for?
        else:
            return points

    def print_geom(self, io: TextIO, indent: str) -> None:
        pass  # Defaults will do for cartesian mesh

    def get_mean_square_distance_weights(self, point):
        ni, nj, nk = self.bins_shape

        def calc_sum(bins):
            bins_square = np.square(bins)
            bins_mult = bins[:-1] * bins[1:]
            bins_square = bins_square[:-1] + bins_square[1:] + bins_mult
            return bins_square

        x_square, y_square, z_square = [
            calc_sum(x - px)
            for x, px in zip((self.ibins, self.jbins, self.kbins), point)
        ]
        w = np.zeros((ni, nj, nk), dtype=float)
        for i in range(ni):
            for j in range(nj):
                for k in range(nk):
                    w[i, j, k] = x_square[i] + y_square[j] + z_square[k]
        w = (1.0 / 3.0) * w

        w = w * (1024.0 / np.max(w))

        return w

    def calc_cell_centers(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} has not implemented method calc_cell_centers"
        )


@dataclass(eq=False)
class CylinderGeometrySpec(AbstractGeometrySpec):
    """Cylinder spec.

    Attributes:
        axs: cylinder axis
        vec: vector to measure angle (theta) from
    """

    axs: np.ndarray = field(default_factory=lambda: DEFAULT_AXIS.copy())
    vec: np.ndarray = field(default_factory=lambda: DEFAULT_VEC.copy())

    def __post_init__(self):
        super().__post_init__()

        if self.axs is not DEFAULT_AXIS:
            if not isinstance(self.axs, np.ndarray):
                raise TypeError(f"Expected axs as numpy array, actual {self.axs}")

        if self.vec is not DEFAULT_VEC:
            if not isinstance(self.vec, np.ndarray):
                raise TypeError(f"Expected vec as numpy array, actual {self.vec}")

        if not (self.theta[0] == 0.0 and self.theta[-1] == 1.0):
            raise ValueError("Theta is expected in rotations only")

    @property
    def bins(self):
        return super().bins + (self.axs, self.vec)

    @property
    def cylinder(self) -> bool:
        return True

    @property
    def r(self) -> np.ndarray:
        return self.ibins

    @property
    def z(self) -> np.ndarray:
        return self.jbins

    @property
    def theta(self) -> np.ndarray:
        return self.kbins

    def local_coordinates(self, points: np.ndarray) -> np.ndarray:
        assert points.shape[-1] == 3, "Expected cartesian point array or single point"
        assert np.array_equal(self.axs, DEFAULT_AXIS) and (
            np.array_equal(self.vec, DEFAULT_VEC)
            or self.vec[1] == 0.0  # vec is in xz plane
        ), "Tilted cylinder meshes are not implemented yet"
        # TODO dvp: implement tilted cylinder meshes
        # ez = self.axs / np.linalg.norm(self.axs)
        # ey = np.cross(ez, self.vec)
        # ey /= np.linalg.norm(ey)
        # ex = np.cross(ey, ez)
        # ex /= np.linalg.norm(ex)
        local_points: np.ndarray = points - self.origin
        local_points[..., :] = (
            np.sqrt(local_points[..., 0] ** 2 + local_points[..., 1] ** 2),  # r
            local_points[..., 2],  # z, just copy
            np.arctan2(local_points[..., 1], local_points[..., 0])
            * _1_TO_2PI,  # theta in rotations
        )
        return local_points

    # TODO dvp: add opposite method global_coordinates

    def print_geom(self, io: TextIO, indent: str) -> None:
        print(indent, "geom=cyl", sep="", file=io)
        print(
            indent,
            "axs=",
            " ".join(format_floats(self.axs)),
            "\n",
            indent,
            "vec=",
            " ".join(format_floats(self.vec)),
            sep="",
            file=io,
        )

    # noinspection SpellCheckingInspection
    def get_mean_square_distance_weights(self, point: np.ndarray) -> np.ndarray:
        ni, nj, nk = self.bins_shape
        assert self.vec is not None
        # Define synonyms for cylinder coordinates
        r = self.ibins  # radius
        phi = self.kbins
        assert phi[-1] == 1.0
        phi = phi * _2PI
        z = self.jbins
        px, py, pz = (
            point - self.origin
        )  # TODO dvp: apply local_coordinates instead of the following
        l1_square = px**2 + py**2
        l1 = np.sqrt(l1_square)  # distance to origin from point projection on z=0 plane
        assert 0.0 < l1
        # Terms of integration of L^2 in cylindrical coordinates
        # r^2
        gamma = np.arcsin(py / l1)
        r_square = np.square(r)
        r_square = 0.5 * (r_square[1:] + r_square[:-1])
        r_sum = r[1:] + r[:-1]
        r_mult = r[1:] * r[:-1]
        dphi = phi[1:] - phi[:-1]
        dsins = np.sin(phi - gamma)
        dsins = dsins[1:] - dsins[:-1]
        dsins = dsins / dphi
        z_minus_pz = z - pz
        z_minus_pz_square = np.square(z_minus_pz)
        z_sum = (1.0 / 3.0) * (
            z_minus_pz_square[1:]
            + z_minus_pz_square[:-1]
            + z_minus_pz[1:] * z_minus_pz[:-1]
        )
        w = np.zeros((ni, nj, nk), dtype=float)

        for i in range(ni):
            for j in range(nj):
                for k in range(nk):
                    a = r_square[i]
                    b = (-4.0 / 3.0) * l1 * (r_sum[i] - r_mult[i] / r_sum[i]) * dsins[k]
                    d = z_sum[j]
                    w[i, j, k] = a + b + d
        w = w + l1_square
        w = w * (1024.0 / np.max(w))

        return w

    def calc_cell_centers(self) -> np.ndarray:
        _x0, _y0, _z0 = self.origin
        r_mids = (self.ibins[1:] + self.ibins[:-1]) * 0.5
        z_mids = (self.jbins[1:] + self.jbins[:-1]) * 0.5
        t_mids = (self.kbins[1:] + self.kbins[:-1]) * 0.5
        if self.kbins[-1] == 1.0:
            t_mids = t_mids * _2PI
        v2 = np.cross(self.axs, self.vec)
        v1 = np.cross(v2, self.axs)
        v2 /= linalg.norm(v2)
        v1 /= linalg.norm(v1)
        axs = self.axs / linalg.norm(self.axs)
        axs_z = np.dot(axs, NZ)

        def _aggregator(elements):
            r, z, fi = elements
            x, y = r * (v1 * np.cos(fi) + v2 * np.sin(fi))[0:2]
            x += _x0
            y += _y0
            z = axs_z * z + _z0
            return np.array([x, y, z], dtype=float)

        cell_centers: np.ndarray = cartesian_product(
            r_mids, z_mids, t_mids, aggregator=_aggregator
        )

        return cell_centers

    # def __hash__(self):
    #     return hash((super().__hash__(), self.axs, self.vec))
    #
    # def __eq__(self, other):
    #     if not isinstance(other, CylinderGeometrySpec):
    #         return False
    #     return (
    #         super().__eq__(other)
    #         and np.array_equal(self.axs, other.axs)
    #         and np.array_equal(self.vec, other.vec)
    #     )

    def adjust_axs_vec_for_mcnp(self) -> "CylinderGeometrySpec":
        """Set `axs` and `vec` attributes to the values, which MCNP considers orthogonal.

        Assumptions
        -----------

        Cylinder mesh is not tilted:
            - `self.vec` is in PY=0 plane
            - `self.axs` is vertical


        Returns
        -------
        gs:
            new CylinderGeometrySpec with adjusted `axs` and `vec` attributes.
        """
        # TODO dvp: fix for arbitrary axs and vec
        axs = self.origin + DEFAULT_AXIS * self.z[-1]
        vec = self.origin + DEFAULT_VEC * self.r[-1]
        return CylinderGeometrySpec(
            self.r, self.z, self.theta, origin=self.origin, axs=axs, vec=vec
        )


def _print_bins(indent, prefix, _ibins, io, columns: int = 6) -> None:
    intervals, coarse_mesh = compute_intervals_and_coarse_bins(_ibins)
    coarse_mesh = coarse_mesh[1:]  # drop the first value - it's presented with origin
    print(indent, f"{prefix}mesh=", sep="", end="", file=io)
    second_indent = indent + " " * 5
    ut.print_n(
        (f"{x:.6g}" for x in coarse_mesh), io=io, indent=second_indent, columns=columns
    )
    print(indent, f"{prefix}ints=", sep="", end="", file=io)
    ut.print_n(intervals, io=io, indent=second_indent, columns=columns)


def select_indexes(
    a: np.ndarray, x: Optional[Union[float, List[float], np.ndarray]]
) -> Union[int, slice, np.ndarray]:
    """Find indexes for a mesh bin, corresponding given coordinates.

    Assumes that `a` is sorted.

    Examples:

        >>> r = np.arange(5)
        >>> r
        array([0, 1, 2, 3, 4])

        For x is None return slice over all `a` indexes.

        >>> select_indexes(r, None)
        slice(0, 5, None)

        For non specified x, if input array represents just one bin,
        then return index 0 to squeeze results.
        >>> select_indexes(np.array([10,20]), None)
        0

        For x = 1.5, we have 1 < 1.5 < 2, so the bin index is to be 1
        >>> select_indexes(r, 1.5)
        1

        For x = 0, it's the first bin, and index is to be 0
        >>> select_indexes(r, 0)
        0

        For coordinates below r[0] return -1.
        >>> select_indexes(r, -1)
        -1

        For coordinates above  r[-1] return a.size-1.
        >>> select_indexes(r, 5)
        4

        And for array of coordinates
        >>> select_indexes(r, np.array([1.5, 0, -1, 5]))
        array([ 1,  0, -1,  4])

    Args:
        a:  bin boundaries
        x: one or more coordinates along `a`-boundaries

    Returns:
        out: index or indices for each given coordinate
    """
    assert 1 < a.size, "Parameter a doesn't represent binning"

    if x is None:
        return slice(0, a.size) if 2 < a.size else 0  # squeeze if there's only one bin

    i: np.ndarray = a.searchsorted(x) - 1

    if np.isscalar(i):
        if i < 0:
            if x == a[0]:
                return 0
    else:
        neg = i < 0
        if np.any(neg):
            eq_to_min = a[0] == x
            i[np.logical_and(neg, eq_to_min)] = 0

    return i


def format_floats(floats: Iterable[float], _format="{:.6g}") -> Iterable[str]:
    yield from map(_format.format, floats)


def compute_intervals_and_coarse_bins(
    arr: Sequence[float], tolerance: float = 1.0e-4
) -> Tuple[List[int], Sequence[float]]:
    """Compute fine intervals and coarse binning.

    Examples:

    Find equidistant bins and report as intervals
    >>> arr = np.array([1, 2, 3, 4], dtype=float)
    >>> arr
    array([1., 2., 3., 4.])
    >>> intervals, coarse = compute_intervals_and_coarse_bins(arr)
    >>> intervals
    [3]
    >>> coarse
    [1.0, 4.0]

    A bins with two interval values.
    >>> arr = np.array([1, 2, 3, 6, 8, 10], dtype=float)
    >>> intervals, coarse = compute_intervals_and_coarse_bins(arr)
    >>> intervals
    [2, 1, 2]
    >>> coarse
    [1.0, 3.0, 6.0, 10.0]

    On zero (or negative tolerance) just use 1 intervals and return original array.
    >>> intervals, coarse = compute_intervals_and_coarse_bins(arr, tolerance=0.0)
    >>> intervals
    [1, 1, 1, 1, 1]
    >>> coarse is arr
    True


    Args:
        arr: actual bins
        tolerance: precision to distinguish intervals with

    Returns:
        Tuple: numbers of fine intervals between coarse bins, coarse binning
    """
    if tolerance <= 0.0:
        return [1] * (len(arr) - 1), arr
    fine_intervals = []
    coarse_bins = [arr[0]]
    d_old = arr[1] - arr[0]
    count = 0
    for i in range(1, len(arr)):
        d = arr[i] - arr[i - 1]
        if abs(d - d_old) < tolerance:
            count += 1
        else:
            d_old = d
            fine_intervals.append(count)
            coarse_bins.append(arr[i - 1])
            count = 1
    fine_intervals.append(count)
    coarse_bins.append(arr[-1])
    return fine_intervals, coarse_bins


def arrays_equal(arrays: Iterable[Tuple[np.ndarray, np.ndarray]]) -> bool:
    for a, b in arrays:
        if not (a is b or np.array_equal(a, b)):
            return False
    return True



"""The CLI application module.

The module applies :meth:`click` API to organize CLI interface for McKit-meshes package.
"""
from __future__ import annotations

import typing as t

import datetime

from pathlib import Path

import click
import mckit_meshes.version as meta

from mckit_meshes.cli.commands import do_mesh2npz, do_npz2vtk

# from mckit_meshes.cli.commands.common import get_default_output_directory
from mckit_meshes.cli.logging import init_logger, logger

# from mckit_meshes.utils import MCNP_ENCODING

NAME = "mckit_meshes"
VERSION = meta.__version__
# LOG_FILE_RETENTION = 3
# NO_LEVEL_BELOW = 30


@click.group("mckit-meshes", help=meta.__summary__)
@click.pass_context
@click.option("--override/--no-override", is_flag=True, default=False)
@click.option(
    "--verbose/--no-verbose", is_flag=True, default=False, help="Log everything"
)
@click.option(
    "--quiet/--no-quiet",
    is_flag=True,
    default=False,
    help="Log only WARNINGS and above",
)
@click.option("--logfile", default=None, help="File to log to")
@click.version_option(VERSION, prog_name=NAME)
def mckit_meshes(
    ctx: click.Context, verbose: bool, quiet: bool, logfile: str, override: bool
) -> None:
    """McKit-meshes command line interface."""
    init_logger(logfile, quiet, verbose)
    # TODO dvp: add customized logger configuring from a configuration toml-file.

    obj = ctx.ensure_object(dict)
    obj["OVERRIDE"] = override


@mckit_meshes.command()
@click.pass_context
@click.option(
    "--prefix",
    "-p",
    default="npz",
    help="""A prefix to prepend output files (default: "npz/"),
output files are also prepended with MESH_TALLY file base name,
if there are more than 1 input file""",
)
@click.argument(
    "mesh_tallies",
    metavar="[<meshtally_file>...]",
    type=click.Path(exists=True),
    nargs=-1,
    required=False,
)
def mesh2npz(
    ctx: click.Context, prefix: str | Path, mesh_tallies: t.List[t.Any]
) -> None:
    """Converts mesh files to npz files."""
    do_mesh2npz(prefix, mesh_tallies, ctx.obj["OVERRIDE"])
    #


@mckit_meshes.command()
@click.pass_context
@click.option(
    "--prefix",
    "-p",
    default=".",
    help="""A prefix to prepend output files (default: "./"),
output files are also prepended with MESH_TALLY file base name,
if there are more than 1 input file""",
)
@click.argument(
    "npz_files",
    metavar="[<npz_file>...]",
    type=click.Path(exists=True),
    nargs=-1,
    required=False,
)
def npz2vtk(ctx: click.Context, prefix: str | Path, npz_files: t.List[t.Any]) -> None:
    """Converts npz files to VTK files."""
    do_npz2vtk(prefix, npz_files, ctx.obj["OVERRIDE"])
    # Don't remove these comments: this makes flake8 happy on absent arguments in the docstring.
    #


if __name__ == "__main__":
    ct = datetime.datetime.now()
    mckit_meshes(obj={})
    logger.success(f"Elapsed time: {datetime.datetime.now() - ct}")



from typing import Any, Callable, Iterable, TextIO

import sys


def foreach(
    predicate: Callable[[Any], Any],
    iterable: Iterable[Any],
) -> None:
    """
    Applies `predicate` to all the items in `iterable`.
    """
    for i in iterable:
        predicate(i)


def print_n(words: Iterable, io: TextIO = None, indent=None, columns: int = 5) -> None:
    if io is None:
        io = sys.stdout
    if indent is None:
        indent = ""
    i = 0
    for w in words:
        if columns <= i:
            print("\n", indent, sep="", end="", file=io)
            i = 0
        if 0 < i:
            print(end=" ", file=io)
        print(w, end="", file=io)
        i = i + 1
    if 0 < i:
        print(file=io)



from typing import Generator, List, TextIO


def m_file_iterator(stream: TextIO) -> Generator[List[str], None, None]:
    header: List[str] = []
    for line in stream:
        if len(header) < 3:
            line = line.strip()
            header.append(line)
        else:
            break
    yield header
    mesh: List[str] = []
    for line in stream:
        line = line.strip()
        if line:
            if len(mesh) > 0 and line.startswith("Mesh Tally Number"):
                yield mesh
                mesh = []
            mesh.append(line)
    if len(mesh) > 0:
        yield mesh



import sys


def mesh2com(xbins, ybins, zbins, out=sys.stdout):
    """
    Prints content for mcnp com file to plot crossections over mesh
    voxels centers and with normals in x, y and z directions.
    """
    print("label 0 0", file=out)
    print("file all", file=out)
    current_basis = current_extent = current_origin = None
    basis = "0 1 0 0 0 1"
    extent = ybins[-1] - ybins[0], zbins[-1] - zbins[0]
    extent = max(*extent)
    extent = extent, extent
    origin_y, origin_z = 0.5 * (ybins[-1] + ybins[0]), 0.5 * (zbins[-1] + zbins[0])
    xmids = 0.5 * (xbins[:-1] + xbins[1:])
    for x in xmids:
        command = ""
        if not current_basis or current_basis != basis:
            command = "basis " + basis
            current_basis = basis
        if not current_extent or current_extent != extent:
            command += " extent {} {}".format(*extent)
            current_extent = extent
        origin = x, origin_y, origin_z
        if not current_origin or current_origin != origin:
            command += " origin {} {} {}".format(*origin)
            current_origin = origin
        if command:
            print(command, file=out)
    basis = "1 0 0 0 0 1"
    current_extent = None
    extent = xbins[-1] - xbins[0], zbins[-1] - zbins[0]
    extent = max(*extent)
    extent = extent, extent
    origin_x, origin_z = 0.5 * (xbins[-1] + xbins[0]), 0.5 * (zbins[-1] + zbins[0])
    ymids = 0.5 * (ybins[:-1] + ybins[1:])
    for y in ymids:
        command = ""
        if not current_basis or current_basis != basis:
            command = "basis " + basis
            current_basis = basis
        if not current_extent or current_extent != extent:
            command += " extent {} {}".format(*extent)
            current_extent = extent
        origin = origin_x, y, origin_z
        if not current_origin or current_origin != origin:
            command += " origin {} {} {}".format(*origin)
            current_origin = origin
        if command:
            print(command, file=out)
    basis = "1 0 0 0 1 0"
    current_extent = None
    extent = xbins[-1] - xbins[0], ybins[-1] - ybins[0]
    extent = max(*extent)
    extent = extent, extent
    origin_x, origin_y = 0.5 * (xbins[-1] + xbins[0]), 0.5 * (ybins[-1] + ybins[0])
    zmids = 0.5 * (zbins[:-1] + zbins[1:])
    for z in zmids:
        command = ""
        if not current_basis or current_basis != basis:
            command = "basis " + basis
            current_basis = basis
        if not current_extent or current_extent != extent:
            command += " extent {} {}".format(*extent)
            current_extent = extent
        origin = origin_x, origin_y, z
        if not current_origin or current_origin != origin:
            command += " origin {} {} {}".format(*origin)
            current_origin = origin
        if command:
            print(command, file=out)
    print("end", file=out)



"""Functions for rebinning histogram-like distributions."""

# TODO: DVP: implement propagation in result the indexes computed on shrink
# for reuse in FMesh.shrink for equivalent grids or alike

from typing import Iterable, Tuple

import collections.abc
import gc
import itertools
import platform

import numpy as np

from numpy import ndarray

if platform.system() == "Linux":
    from mckit_meshes.utils.no_daemon_process import Pool
else:
    Pool = None

# TODO

__all__ = [
    "interpolate",
    "is_monotonically_increasing",
    "rebin_1d",
    "rebin_nd",
    "rebin_spec_composer",
    "shrink_1d",
    "shrink_nd",
    "trim_spec_composer",
]

__revision__ = "$Id$"

__ZERO = np.array([0.0], dtype=float)
__EXTERNAL_PROCESS_THRESHOLD = 1000000


# noinspection PyTypeChecker
def is_monotonically_increasing(a: ndarray):
    if not a.size:
        return False
    iterator = iter(a)
    prev = next(iterator)
    for val in iterator:
        if prev < val:
            prev = val
        else:
            return False
    return True


def set_axis(indices, axis, a_shape):
    shape = [1] * len(a_shape)
    shape[axis] = a_shape[axis]
    return indices.reshape(tuple(shape))


def interpolate(x_new, x, y, axis=None):
    if y.ndim == 1:
        return np.interp(x_new, x, y)

    if axis is None:
        axis = 0

    x_new_indices = np.digitize(x_new, x)
    x_new_indices = x_new_indices.clip(1, len(x) - 1).astype(int)
    lo = x_new_indices - 1
    hi = x_new_indices
    x_lo = x[lo]
    deltas = x[hi] - x_lo
    nd = y.ndim
    slice1 = [slice(None)] * nd
    slice2 = [slice(None)] * nd
    slice1[axis] = lo
    slice2[axis] = hi
    slice1 = tuple(slice1)
    slice2 = tuple(slice2)
    y_lo = y[slice1]
    y_deltas = y[slice2] - y_lo
    deltas = set_axis(deltas, axis, y_deltas.shape)
    slope = y_deltas / deltas
    new_deltas = x_new - x_lo
    new_deltas = set_axis(new_deltas, axis, slope.shape)
    y_new = slope * new_deltas + y_lo
    return y_new


def rebin_1d(a, bins, new_bins, axis=0, grouped=False, assume_sorted=False):
    """Transforms 1-D histogram defined as `data` on the limiting points
    define like `bins` to equivalent (see the terms below) histogram defined
    on other limiting points defined as `new_bins`.

    Notes
    -----
    The algorithm maintains the equality of integral on intervals defined on
    new_bins for the original and rebinned distributions.

    Parameters
    ----------
        a: ndarray
            The array to rebin
        bins: ndarray
            Defines 1-D array representing `a` binning along the given `axis
        new_bins:  ndarray
            The new binning required.
        axis: int, optional
            An axis along which to rebin array `a`
        grouped: bool, optional
            Defines the approach for rebinning.

            - If `True`, then the values in `a` represent the data already
               integrated over bins, like in energy group distributions.
               On rebinning maintain equivalence of integral over same
               energy range in old and new bins.
            - If `False` (default), as for spatial binning - maintain the same
              average value in the same volume in old and new bins.
        assume_sorted: bool, optional
            If True, then skip assertion of bins sorting order,
            by default False - asserts the input_file data

    Returns
    -------
        rebinned_data: ndarray
    """
    assert (
        bins[0] <= new_bins[0] and new_bins[-1] <= bins[-1]
    ), "Rebinning doesn't provide extrapolation out of the original bins"
    assert (
        bins.size == a.shape[axis] + 1
    ), "The `a` array shape doesn't match the given bins and axis"
    assert assume_sorted or is_monotonically_increasing(bins)
    assert assume_sorted or is_monotonically_increasing(new_bins)

    ndim = a.ndim

    if grouped:
        t = a
    else:
        diff = np.diff(bins)
        if ndim > 1:
            diffs_shape = [1] * ndim
            diffs_shape[axis] = a.shape[axis]
            diffs_shape = tuple(diffs_shape)
            diff = diff.reshape(diffs_shape)
        t = a * diff

    cum = np.cumsum(t, axis=axis)
    cum = np.insert(cum, 0, __ZERO, axis=axis)
    rebinned_data = interpolate(new_bins, bins, cum, axis=axis)
    rebinned_data = np.diff(rebinned_data, axis=axis)

    # del cum
    if a.size > __EXTERNAL_PROCESS_THRESHOLD:
        gc.collect()
        del gc.garbage[:]

    if not grouped:
        new_diff = np.diff(new_bins)
        if ndim > 1:
            diffs_shape = [1] * ndim
            diffs_shape[axis] = rebinned_data.shape[axis]
            new_diff = new_diff.reshape(tuple(diffs_shape))
        rebinned_data /= new_diff

    return rebinned_data


def rebin_nd(
    a: ndarray,
    rebin_spec: Iterable[Tuple[ndarray, ndarray, int, bool]],
    assume_sorted: bool = False,
    external_process_threshold: int = __EXTERNAL_PROCESS_THRESHOLD,
) -> ndarray:
    """
    Rebin an array `a` over multidimensional grid.

    Args:
        a: An array to rebin.
        rebin_spec: Iterator
            An iterator listing tuples specifying  bins, new_bins, axis and
            grouped  parameters for rebinning.
            See :py:func:`rebin_1d` for details on the parameters.
        assume_sorted: bool, optional
            If True skip assertion of bins sorting order,
            by default False - asserts the input_file data
        external_process_threshold: int
            If size of `a` is greater than that, then the computation is executed in external process,
            to achieve immediate memory cleanup.
    Returns
    --------
        rebinned_data: ndarray
    """
    if not isinstance(rebin_spec, collections.abc.Iterator):
        rebin_spec = iter(rebin_spec)
    try:
        bins, new_bins, axis, grouped = next(rebin_spec)
    except StopIteration:
        return a

    if Pool is not None and a.size > external_process_threshold:
        recursion_res = Pool(processes=1).apply(
            rebin_nd, args=(a, rebin_spec, assume_sorted)
        )
    else:
        recursion_res = rebin_nd(a, rebin_spec, assume_sorted)

    res = rebin_1d(recursion_res, bins, new_bins, axis, grouped, assume_sorted)

    del recursion_res
    if a.size > external_process_threshold:
        n = gc.collect()
        if n:
            del gc.garbage[:]

    return res


def rebin_spec_composer(bins_seq, new_bins_seq, axes=None, grouped_flags=None):
    """Compose rebin_spec parameter.

    See also :py:func:`mckit_meshes.utils.rebin.rebin_nd` with reasonable defaults
    for axes and grouped iterators.

    Args:
        bins_seq: sequence of ndarrays
            Iterates over the list of original bins
        new_bins_seq: sequence of ndarrays
            Iterates over the list of new bins.
        axes: sequence of ints, optional
            Iterates over the list of corresponding axes.
            If not provided (default), then iterates over sequence 0 ... len(bins).
        grouped_flags: sequence of booleans, optional
            Iterates over a sequence of grouped flags.
            If not provided (default),
            then all the axes considered as not grouped.
            If constant boolean value is proved, then for
            all the axes this value is applied.

    Returns:
        Iterator: ... over the sequence of tuples (bins, new_bins, axis, grouped)
    """
    if not axes:
        axes = itertools.count()
    if isinstance(grouped_flags, bool):
        grouped_flags = itertools.repeat(grouped_flags)
    elif not grouped_flags:
        grouped_flags = itertools.repeat(False)
    return zip(bins_seq, new_bins_seq, axes, grouped_flags)


# @numba.jit
def shrink_1d(
    a: np.ndarray, bins: np.ndarray, low=None, high=None, axis=None, assume_sorted=False
) -> Tuple[np.ndarray, np.ndarray]:
    """Select sub-arrays of a `a` and corresponding `bins` for minimal span
    of bins, which completely covers the range [`low`..`high`]
    both sides included.

    Args:
        a: ndarray
            An array to shrink.
        bins: ndarray
            Bins corresponding to the grid `a` over the given `axis`.
        low: float, optional
            Left edge of the range to shrink to.
            When omitted, the `bins` left edge is used.
        high: float
            Right edge of the range to shrink to.
            When omitted, the `bins` right edge is used.
        axis: int, optional
            An axis of `a` over which to shrink. Default axis = 0.
        assume_sorted: bool, optional
            If True skip assertion of bins sorting order,
            by default False - asserts the input_file data

    Returns:
        new_bins: ndarray
            The shrank bins
        new_data: ndarray
            The shrank grid

    """

    if low is None and high is None:
        return bins, a

    if axis is None:
        axis = 0

    assert a.shape[axis] == bins.size - 1
    assert assume_sorted or is_monotonically_increasing(bins)

    if low is None:
        low = bins[0]

    if high is None:
        high = bins[-1]

    if low == bins[0] and high == bins[-1]:
        return bins, a

    if low < bins[0] or bins[-1] < low:
        raise ValueError(
            "Low shrink edge is beyond the bins range: %g is not in [%g..%g]"
            % (low, bins[0], bins[-1])
        )

    if high < bins[0] or bins[-1] < high:
        raise ValueError(
            "High shrink edge is beyond the bins range: %g is not in [%g..%g]"
            % (high, bins[1], bins[-1])
        )

    left_idx, right_idx = np.digitize([low, high], bins) - 1

    if left_idx > 0 and bins[left_idx] > low:
        left_idx -= 1

    if right_idx < bins.size - 1 and bins[right_idx] < high:
        right_idx += 1

    if right_idx - left_idx < 1:
        raise ValueError("Shrink results to empty grid")

    if left_idx == 0 and right_idx == bins.size:
        return bins, a

    indices = list(range(left_idx, right_idx + 1))
    new_bins = np.take(bins, indices)
    new_a = np.take(a, indices[:-1], axis=axis)

    return new_bins, new_a


def shrink_nd(a, trim_spec, assume_sorted=False):
    """Perform multidimensional shrink.

    Args:
        a: ndarray
            The grid to shrink.
        trim_spec: sequence of tuples
            Iterates over tuples (bins, low, high, axis)
        assume_sorted: bool, optional
            If True skip assertion of bins sorting order,
            by default False - asserts the input_file data

    Returns:
        new_bins_seq: sequence of ndarrays
            A sequence with  new bins.
        recursed_data: ndarray
            The shrunk grid.
    """
    if not isinstance(trim_spec, collections.abc.Iterator):
        trim_spec = iter(trim_spec)
    try:
        bins, left, right, axis = next(trim_spec)
    except StopIteration:
        return None, a
    new_bins_seq, recursed_data = shrink_nd(a, trim_spec, assume_sorted)
    top_bins, top_data = shrink_1d(
        recursed_data, bins, left, right, axis, assume_sorted
    )
    if new_bins_seq:
        new_bins_seq = [top_bins] + new_bins_seq
    else:
        new_bins_seq = [top_bins]
    return new_bins_seq, top_data


def trim_spec_composer(bins_seq, lefts=None, rights=None, axes=None):
    """
    Helps to compose trim_spec parameter in
    :py:func:`triniti_ne.rebin.trim_nd` with
    reasonable defaults for lefts, rights and axes iterators.

    Parameters
    ----------
    bins_seq: sequence of ndarrays
        Iterates over the list of bins associated with a grid to be trimmed.
    lefts: sequence of floats
        Iterates over the list of left edges for trimming.
    rights: sequence of floats
        Iterates over the list of right edges for trimming.
    axes: sequence of ints, optional
        Iterates over the list of corresponding axes.
        If not provided (default), then iterates over sequence 0..len(bins).

    Returns
    -------
    trim_spec: sequence of tuples
        Iterator over the sequence of tuples (bins, lefts, rights, axis)
    """
    if not lefts:
        lefts = itertools.repeat(None)
    if not rights:
        rights = itertools.repeat(None)
    if not axes:
        axes = itertools.count()
    return zip(bins_seq, lefts, rights, axes)



from enum import IntEnum


class ParticleKind(IntEnum):
    neutron = 1
    photon = 2
    electron = 3
    n = 1
    p = 2
    e = 3

    @property
    def short(self) -> str:
        return self.name[0]

    @property
    def heating_reactions(self) -> str:
        if self is self.n:
            return "1 -4"
        elif self is self.p:
            return "-5 -6"
        else:
            raise RuntimeError("Heating spec is not defined for " + self.name)



"""Output utilities."""
from __future__ import annotations

import typing as t

from typing import Any, Iterable, TextIO

import sys

from pathlib import Path


def print_cols(
    seq: Iterable[Any], fid: TextIO = sys.stdout, max_columns: int = 6, fmt: str = "{}"
) -> int:
    """Print sequence in columns.

    Args:
        seq: sequence to print
        fid: output
        max_columns: max columns in a line
        fmt: format string

    Returns:
        int: the number of the last column printed on the last row

    """
    column = 0
    for s in seq:
        print(fmt.format(s), file=fid, end="")
        column += 1
        if column == max_columns:
            print(file=fid)
            column = 0
        else:
            print(" ", file=fid, end="")
    return column


def ignore_existing_file_strategy(_: str | Path) -> None:
    pass


def raise_error_when_file_exists_strategy(path: str | Path) -> None:
    path = Path(path)
    if path.exists():
        errmsg = f"""\
Cannot override existing file \"{path}\".
Please remove the file or specify --override option"""
        # raise click.UsageError(errmsg)
        raise FileExistsError(errmsg)


def check_if_path_exists(override: bool) -> t.Callable[[str | Path], None]:
    return (
        ignore_existing_file_strategy
        if override
        else raise_error_when_file_exists_strategy
    )



from typing import cast

import numpy as np

__all__ = ["a"]


def a(*args, dtype=float) -> np.ndarray:
    """Shorten typing in parametrized tests.

    Equivalent to np.ndarray([*args], dtype='dtype')`.

    Args:
        args:  sequence of numbers (any type) will be converted to the specified type.
        dtype: A type for the output array

    Returns:
        np.ndarray:  The ndarray with the given numbers and type.

    Examples:

    >>> a(1, 2, 3)
    array([1., 2., 3.])
    >>> a(1, 2, 3, dtype=np.int32)
    array([1, 2, 3], dtype=int32)
    >>> a(0, 1, 2, 3).reshape(2,2)
    array([[0., 1.],
           [2., 3.]])
    """
    return cast(np.ndarray, np.fromiter(args, dtype=dtype))


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



"""Classes and functions for operations with fmesh tallies.

Classes
-------
FMesh - class for storing individual fmesh tally data.

Functions
---------
read_meshtal(file) - reads all tallies from MCNP meshtal file.
merge_tallies(*tally_weight) - merges tallies with specific weights.

Examples
--------
Suppose we have meshtal file - MCNP output for fmesh tallies - sample.m.
It contains several tallies with numbers 14, 24, and 34. First, we have to
read it:

    ```python
        tally_list = list(read_meshtal('sample.m'))
    ```

x and z are coordinates of mesh cell centers along x and z axis. data is
2-dimensional array of fmesh tally values. err - relative errors.

In order to save fmesh data to vtk format, method save2vtk should be used.
Because it is actually regular grid the result is .vtr file.

    ```python
        tally_list[0].save2vtk(filename='sample', data_name='heating neutron')
    ```

Dependencies
------------

pyevkt
    https://github.com/pyscience-projects/pyevtk
    https://bitbucket.org/pauloh/pyevtk
"""

# TODO dvp: redesign this class as xarray data structure.
#           multidimensional array with coordinates is more appropriate for this class.

from __future__ import annotations

from typing import Callable, Generator, Iterable, List, Optional, TextIO, Tuple, Union

import logging

from pathlib import Path

import mckit_meshes.mesh.geometry_spec as gc
import mckit_meshes.utils.no_daemon_process as ndp
import mckit_meshes.utils.rebin as rebin
import numpy as np

from mckit_meshes.particle_kind import ParticleKind as Kind
from mckit_meshes.utils.io import raise_error_when_file_exists_strategy
from pyevtk.hl import gridToVTK
from toolz.itertoolz import concatv

__LOG = logging.getLogger(__name__)


def _expand_args(args):
    return rebin.rebin_nd(*args)


class FMesh:
    """Fmesh tally object.

    Attributes:

    name: The number of tally.

    kind: Kind of tally: neutron, photon, electron or generic.
    """

    NPZ_MARK = np.int16(5445)
    """
    'Signature' to be stored in the first entry of meta entry in an npz file to check that the file is for FMesh object
    """

    NPZ_FORMAT = np.int16(4)
    """
    Identifies version of format of data stored in npz file
    """

    class X(RuntimeError):
        pass

    def __init__(
        self,
        name: int,
        kind: int | Kind,
        geometry_spec: gc.AbstractGeometrySpec,
        ebins: np.ndarray,
        data: np.ndarray,
        errors: np.ndarray,
        totals: Optional[np.ndarray] = None,
        totals_err: Optional[np.ndarray] = None,
        comment: Optional[str] = None,
    ) -> None:
        """Construct FMesh instance object.

        Args:
            name: 'name' attribute value
            kind: 'kind' attribute value
            geometry_spec: mesh geometry specification
            ebins: Energy bin boundaries.

            data: Data values at centers of mesh cells.
                Shape (Ne-1)x(Nx-1)x(Ny-1)x(Nz-1), where Ne, Nx, Ny and Nz - the number
                of corresponding bin boundaries.

            errors:
                Relative errors of corresponding data values.
                Shape (Ne-1)x(Nx-1)x(Ny-1)x(Nz-1), where Ne, Nx, Ny and Nz - the number
                of corresponding bin boundaries.
            totals: Can be provided with 'Total' data from mesh file if there are more than 1 energy bin, optional.
            totals_err: Can be provided with data from mesh file if there are more than 1 energy bin, optional.
            comment: Comment from a meshtal file (content of FC card in MCNP model).
        """
        self.name = int(name)
        self.kind = Kind(kind)

        self._geometry_spec: Union[
            gc.CartesianGeometrySpec, gc.CylinderGeometrySpec, gc.AbstractGeometrySpec
        ] = geometry_spec
        self.bins = {}
        self.bins["X"] = self._x = geometry_spec.ibins
        self.bins["Y"] = self._y = geometry_spec.jbins
        self.bins["Z"] = self._z = geometry_spec.kbins
        self.bins["E"] = self._e = gc.as_float_array(ebins)
        self.data = gc.as_float_array(data)
        self.errors = gc.as_float_array(errors)
        # self._totals = totals
        # self._totals_err = totals_err
        if 2 < self._e.size:
            if totals is None or totals_err is None:
                assert (
                    totals is None and totals_err is None
                ), "Both totals and totals_err are to be provided or omitted"
                self._totals = np.sum(self.data, axis=0)
                non_zero = self._totals > 0.0
                self._totals_err = np.zeros_like(self._totals)
                self._totals_err[non_zero] = (
                    np.sqrt(np.sum((self.errors * self.data) ** 2, axis=0))[non_zero]
                    / self._totals[non_zero]
                )
            else:
                assert (
                    totals is not None and totals_err is not None
                ), "Both totals and totals_err are to be provided or omitted"
                self._totals = np.asarray(totals, dtype=float)
                self._totals_err = np.asarray(totals_err, dtype=float)
        else:
            self._totals = None
            self._totals_err = None
        self._comment = comment
        self.check_attributes()

    def check_attributes(self):
        """Check if attributes shapes correspond to  each other."""
        assert 2 <= self._e.size
        assert self.data.shape == self.errors.shape
        assert self.data.shape == (self.e.size - 1,) + self._geometry_spec.bins_shape
        assert (
            self._totals is None
            or isinstance(self._totals, np.ndarray)
            and isinstance(self._totals_err, np.ndarray)
            and self._totals.shape == self._totals_err.shape
            and self._totals.shape == self._geometry_spec.bins_shape
        )

    # @property
    # def x(self):
    #     return self._x
    #
    # @property
    # def y(self):
    #     return self._y
    #
    # @property
    # def z(self):
    #     return self._z

    @property
    def e(self) -> np.ndarray:
        """Energy bins."""
        return self._e

    @property
    def has_multiple_energy_bins(self) -> bool:
        """Check if there's more than 1 energy bin.

        If True, then totals and totals err should present.
        """
        return 2 < self.e.size

    @property
    def ibins(self) -> np.ndarray:
        """Synonym to geometry ibins (x or R)."""
        return self._geometry_spec.ibins

    @property
    def jbins(self) -> np.ndarray:
        """Synonym to geometry jbins (y or Z)."""
        return self._geometry_spec.jbins

    @property
    def kbins(self) -> np.ndarray:
        """Synonym to geometry kbins (z or Theta)."""
        return self._geometry_spec.kbins

    @property
    def totals(self) -> np.ndarray | None:
        """Total values over energy."""
        return self._totals

    @property
    def totals_err(self) -> np.ndarray | None:
        """Relative errors of total values over energy."""
        return self._totals_err

    @property
    def comment(self) -> str | None:
        """Comment from FC card for this mesh tally."""
        return self._comment

    @property
    def origin(self):
        assert self.is_cylinder, "Only valid for cylinder mesh."
        return self._geometry_spec.origin

    @property
    def axis(self) -> np.ndarray:
        assert self.is_cylinder, "Only valid for cylinder mesh."
        return self._geometry_spec.axs

    @property
    def vec(self) -> np.ndarray:
        assert self.is_cylinder, "Only valid for cylinder mesh."
        return self._geometry_spec.vec

    @property
    def is_cylinder(self) -> bool:
        """Is this mesh cylinder?

        Note:
            MCNP uses `origin` on mesh tally specification, both rectilinear and cylinder,
            but outputs origin only for cylinder mesh.
        """
        return self._geometry_spec.cylinder

    @property
    def total_precision(self) -> float:
        if self.has_multiple_energy_bins:
            return self.totals_err[
                -1
            ]  # TODO dvp: assumes max energy bin is most representative, check usage
        else:
            return self.errors[0, 0, 0, 0]

    def is_equal_by_mesh(self, other: "FMesh") -> bool:
        return (
            self.kind == other.kind
            and self._geometry_spec == other._geometry_spec
            and np.array_equal(self.e, other.e)
        )

    def has_better_precision_than(self, other) -> bool:
        assert self.is_equal_by_mesh(
            other
        ), "Incompatible meshes for precision comparison."
        return self.total_precision < other.total_precision

    def __eq__(self, other) -> bool:
        if not isinstance(other, FMesh):
            return False
        res = (
            self.name == other.name
            and self.is_equal_by_mesh(other)
            and np.array_equal(self.data, other.data)
            and np.array_equal(self.errors, other.errors)
            and self.comment == other.comment
        )
        if res and self._totals:
            res = np.all(np.isclose(self.totals, other.totals)) and np.all(
                np.isclose(self.totals_err, other.totals_err)
            )
        # if res:
        #     res = self.is_cylinder == other.is_cylinder
        #     if res and self.is_cylinder:
        #         res = not self.origin(
        #             np.all(np.isclose(self.origin, other.origin))
        #             and np.all(np.isclose(self.axis, other.axis))
        #         )
        return res

    def __hash__(self) -> int:
        return hash(
            (
                self.name,
                self.kind,
                self._geometry_spec,
                self.e,
                self.data,
                self.errors,
                self.comment,
            )
        )

    def __repr__(self) -> str:
        msg = "Fmesh({name}, {kind}, {xmin}..{xmax}, {ymin}..{ymax}, {zmin}..{zmax}, {emin}..{emax})"
        (xmin, xmax), (ymin, ymax), (zmin, zmax) = self._geometry_spec.boundaries
        return msg.format(
            name=self.name,
            kind=self.kind,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            zmin=zmin,
            zmax=zmax,
            emin=self.e[0],
            emax=self.e[-1],
        )

    def surrounds_point(self, x: float, y: float, z: float, local: bool = True) -> bool:
        return self._geometry_spec.surrounds_point(x, y, z, local)

    # def get_slice(
    #     self, free: str = "XY", **kwargs
    # ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    #     """Gets slice of fmesh tally.
    #
    #     kwargs specify names of fixed variables and their values. If the key
    #     corresponding to the one of fixed variables is present, then its value
    #     is ignored. If the point is outside the mesh then zeros are returned.
    #
    #     Parameters
    #     ----------
    #     free : str
    #         Names of free parameters (those, which corresponds to slice axis).
    #     kwargs : dict
    #         Key-value pairs of fixed parameters. Possible keys:
    #         E - energy value of slice. If 'total' - data is summed over all
    #         energy bins; default: 'total';
    #         X - position of slice in x direction; default: first bin;
    #         Y - position of slice in y direction; default: first bin;
    #         Z - position of slice in z direction; default: first bin;
    #
    #     Returns
    #     -------
    #     x_centers, y_centers : numpy.ndarray, ndim=1
    #         Coordinates of cell centers along free variables.
    #     result, error : numpy.ndarray, ndim=2
    #         An array of data values in the specified section (in phase space).
    #     """
    #     key_index = {0: "E", 1: "X", 2: "Y", 3: "Z"}
    #     free = free.upper()
    #     free_keys = [free[0], free[1]]
    #     slice_index = []
    #     sum_axis = []
    #     for i in range(4):
    #         key = key_index[i]
    #         if key in free_keys:
    #             index = np.arange(self.bins[key].size - 1)
    #         elif key in kwargs and isinstance(kwargs[key], (int, float)):
    #             index = np.searchsorted(self.bins[key], kwargs[key]) - 1
    #         elif key == "E":
    #             index = np.arange(self.bins[key].size - 1)
    #             sum_axis.append(i)
    #         else:
    #             index = 0
    #         slice_index.append(index)
    #
    #     result_data = self.data
    #     result_error = self.errors
    #     for i, index in reversed(list(enumerate(slice_index))):
    #         if not isinstance(index, np.ndarray) and (
    #             index < 0 or index >= self.bins[key_index[i]].size - 1
    #         ):
    #             result_data *= 0
    #             result_error *= 0
    #             index = 0
    #         result_data = result_data.take(index, axis=i)
    #         result_error = result_error.take(index, axis=i)
    #
    #     if sum_axis:
    #         abs_err_square = (result_data * result_error) ** 2
    #         abs_tot_err = np.sqrt(np.sum(abs_err_square, axis=tuple(sum_axis)))
    #         result_data = np.sum(result_data, axis=tuple(sum_axis))
    #         result_error = np.nan_to_num(abs_tot_err / result_data)
    #
    #     xaxs = (self.bins[free_keys[0]][1:] + self.bins[free_keys[0]][:-1]) / 2.0
    #     yaxs = (self.bins[free_keys[1]][1:] + self.bins[free_keys[1]][:-1]) / 2.0
    #
    #     return xaxs, yaxs, result_data, result_error

    def get_spectrum(self, x, y, z):
        """Gets energy spectrum at the specified point.


        Args:
            x, y, z : double
                X, Y and Z coordinate of the point where energy spectrum is
                required. If point is located outside the mesh, zeros are returned.

        Returns:
            ebins, spec, err : numpy.ndarray[double]
                Energy bin boundaries, group energy spectrum and relative errors.
        """
        key_index = {0: "X", 1: "Y", 2: "Z"}
        values = [x, y, z]
        result_data = self.data
        result_error = self.errors
        for i, value in reversed(list(enumerate(values))):
            key = key_index[i]
            index = np.searchsorted(self.bins[key], value) - 1
            if index < 0 or index >= self.bins[key].size - 1:
                result_data *= 0
                result_error *= 0
                index = 0
            result_data = result_data.take(index, axis=i + 1)
            result_error = result_error.take(index, axis=i + 1)
        return self.e, result_data, result_error

    def select_indexes(self, *, x=None, y=None, z=None):
        return self._geometry_spec.select_indexes(i_values=x, j_values=y, k_values=z)

    def get_totals(self, *, x=None, y=None, z=None):
        if self._totals is None:
            return None, None
        found_x, found_y, found_z = self.select_indexes(x=x, y=y, z=z)
        totals, rel_error = (
            self._totals[found_x, found_y, found_z],
            self._totals_err[found_x, found_y, found_z],
        )
        return totals, rel_error

    def save_2_npz(
        self,
        filename: Path,
        check_existing_file_strategy=raise_error_when_file_exists_strategy,
    ) -> None:
        """Writes this object to numpy npz file_.

        Args:
            filename:
                Filename to which the object is saved. If file_ is a
                file-object, then the filename is unchanged. If file_ is a string,
                a .npz extension will be appended to the file_ name if it does not
                already have one. By default, the name of file_ is the tally name.
            check_existing_file_strategy: what to do if an output file already exists
        """
        if not filename.suffix == ".npz":
            filename = filename.with_suffix(".npz")

        check_existing_file_strategy(filename)

        kwd = dict(
            meta=np.array(
                [FMesh.NPZ_MARK, FMesh.NPZ_FORMAT, self.name, self.kind],
                dtype=np.uint32,
            ),
            E=self.e,
            X=self.ibins,
            Y=self.jbins,
            Z=self.kbins,
            data=self.data,
            errors=self.errors,
            totals=self.totals,
            totals_err=self.totals_err,
        )
        if self.comment:
            kwd["comment"] = np.array(self.comment)
        if self.is_cylinder:
            kwd["origin"] = np.array(self._geometry_spec.origin)
            kwd["axis"] = np.array(self._geometry_spec.axs)

        filename.parent.mkdir(parents=True, exist_ok=True)
        np.savez_compressed(str(filename), **kwd)

    @classmethod
    def load_npz(cls, file_: Union[str, Path]) -> "FMesh":
        """

        Loads Fmesh object from the binary file_.

        Parameters
        ----------
        file_ : file or str
            File or filename from which the object will be loaded.
        """
        if isinstance(file_, Path):
            file_ = str(file_)
        with np.load(file_) as data:
            meta = data["meta"]
            mark = meta[0]
            assert mark == FMesh.NPZ_MARK, "Incompatible file format %s" % file_
            version = meta[1]
            name, kind = meta[2:4]
            if 1 <= version:
                e = data["E"]
                x = data["X"]
                y = data["Y"]
                z = data["Z"]
                d = data["data"]
                r = data["errors"]
                if 2 < e.size:
                    try:
                        totals = data["totals"]
                        totals_err = data["totals_err"]
                    except KeyError:
                        totals = None
                        totals_err = None
                else:
                    totals = None
                    totals_err = None
                comment = None
                origin = None
                axis = None
                if 2 <= version:
                    if "comment" in data:
                        comment = data["comment"]
                        comment = comment.item()
                        assert comment
                    if 3 <= version:
                        if "origin" in data:
                            assert "axis" in data
                            origin = data["origin"]
                            axis = data["axis"]
                            assert origin.size == 3
                            assert axis.size == 3
                        if 4 <= version:
                            pass
                        else:
                            kind = int(kind) + 1
                if origin is None:
                    geometry_spec = gc.CartesianGeometrySpec(x, y, z)
                else:
                    geometry_spec = gc.CylinderGeometrySpec(
                        x, y, z, origin=origin, axs=axis
                    )
                return cls(
                    name,
                    kind,
                    geometry_spec,
                    e,
                    d,
                    r,
                    totals,
                    totals_err,
                    comment=comment,
                )
            else:
                raise FMesh.X("Invalid version for FMesh file %d" % version)

    def save2vtk(self, filename: str = None, data_name: str = None) -> str:
        """Saves this fmesh data to vtk file.

        Data is saved for every energy bin and, if there are multiple energy bins,
        for total values (sum across energy axis).

        Args:
            filename :
                Name of file to which this object is stored. A .vtk extension will
                be appended. By default, the name of file is the tally name.
            data_name :
                Name of data which will appear in vtk file. If None, tally name
                and type will be used.
        """
        assert not self.is_cylinder, "Not implemented for cylinder geometry"
        # TODO dvp: implement for cylinder geometry (see iwwgvr project for example).

        if filename is None:
            filename = str(self.name)
        if data_name is None:
            data_name = str(self.name) + " " + self.kind.name

        cell_data = {}
        for i, e in enumerate(self.e[1:]):
            key = data_name + " E={0:.4e}".format(e)
            cell_data[key] = self.data[i, :, :, :]
        if self.has_multiple_energy_bins:
            name = data_name + " total"
            cell_data[name] = np.sum(self.data, axis=0)
        return gridToVTK(
            filename, self.ibins, self.jbins, self.kbins, cellData=cell_data
        )

    # noinspection PyUnresolvedReferences
    def save_2_mcnp_mesh(self, stream: TextIO) -> None:
        """
        Saves the mesh in a file in a format similar to mcnp mesh tally textual representation.

        Args:
            stream :  stream to store the mesh.
        """

        def format_comment(a):
            return "\n" + a.comment if a.comment else ""

        header = f"""
 Mesh Tally Number   {self.name}{format_comment(self)}
 This is a {self.kind.name} mesh tally.

 Tally bin boundaries:{self.format_cylinder_origin_and_axis_label()}
"""[
            1:-1
        ]
        e = self.e[1:]
        x = 0.5 * (self.ibins[1:] + self.ibins[:-1])
        y = 0.5 * (self.jbins[1:] + self.jbins[:-1])
        z = 0.5 * (self.kbins[1:] + self.kbins[:-1])
        print(header, file=stream)
        print(
            f"{'R' if self.is_cylinder else 'X'} direction:",
            file=stream,
            end="",
        )
        for f in np.nditer(self.ibins):
            print("% g" % f, file=stream, end="")
        print(file=stream)
        print(
            "{} direction:".format("Z" if self.is_cylinder else "Y"),
            file=stream,
            end="",
        )
        for f in np.nditer(self.jbins):
            print(" %g" % f, file=stream, end="")
        print(file=stream)
        print(
            "{} direction:".format("Theta" if self.is_cylinder else "Z"),
            file=stream,
            end="",
        )
        for f in np.nditer(self.kbins):
            print(" %g" % f, file=stream, end="")
        print(file=stream)
        print("Energy bin boundaries:", file=stream, end="")
        for f in np.nditer(self.e):
            print(" %g" % f, file=stream, end="")
        print("\n", file=stream)
        if self.is_cylinder:
            print(
                "   Energy         R         Z         Th    Result     Rel Error",
                file=stream,
            )
        else:
            print(
                "   Energy         X         Y         Z     Result     Rel Error",
                file=stream,
            )

        for ie in range(e.size):
            for ix in range(x.size):
                for iy in range(y.size):
                    for iz in range(z.size):
                        value = self.data[ie, ix, iy, iz]
                        err = self.errors[ie, ix, iy, iz]
                        row = " %10.3e%10.3f%10.3f%10.3f %11.5e %11.5e" % (
                            e[ie],
                            x[ix],
                            y[iy],
                            z[iz],
                            value,
                            err,
                        )
                        print(row, file=stream)

        for ix in range(x.size):
            for iy in range(y.size):
                for iz in range(z.size):
                    if self._totals:
                        value = self._totals[ix, iy, iz]
                        err = self._totals_err[ix, iy, iz]
                    else:
                        portion = self.data[:, ix, iy, iz]
                        value = np.sum(portion)
                        err = portion * self.errors[:, ix, iy, iz]
                        err = np.sqrt(np.sum(err * err)) / value
                    row = "%11s%10.3f%10.3f%10.3f %11.5e %11.5e" % (
                        "   Total   ",
                        x[ix],
                        y[iy],
                        z[iz],
                        value,
                        err,
                    )
                    print(row, file=stream, end="")

            print("\n", file=stream)

    def total_by_energy(self, new_name: int = 0) -> "FMesh":
        e = np.array([self.e[0], self.e[-1]])
        data = self.totals[np.newaxis, ...]
        errors = self.totals_err[np.newaxis, ...]
        return FMesh(new_name, self.kind, self._geometry_spec, e, data, errors)

    def shrink(
        self,
        emin=None,
        emax=None,
        xmin=None,
        xmax=None,
        ymin=None,
        ymax=None,
        zmin=None,
        zmax=None,
        new_name=-1,
    ) -> "FMesh":
        """Select subset of e-voxels within given geometry limits.

        Args:
            emin:
            emax:
            xmin:
            xmax:
            ymin:
            ymax:
            zmin:
            zmax:
            new_name: name for mesh to be created, default -1.

        Returns:

        """
        trim_spec = [
            f
            for f in rebin.trim_spec_composer(
                [self.e, self.ibins, self.jbins, self.kbins],
                [emin, xmin, ymin, zmin],
                [emax, xmax, ymax, zmax],
            )
        ]
        new_bins_list, new_data = rebin.shrink_nd(
            self.data, iter(trim_spec), assume_sorted=True
        )
        _, new_errors = rebin.shrink_nd(
            self.errors, iter(trim_spec), assume_sorted=True
        )

        assert all(np.array_equal(a, b) for a, b in zip(new_bins_list, _))

        new_ebins, new_xbins, new_ybins, new_zbins = new_bins_list
        if self.totals is None:
            new_totals = None
            new_totals_err = None
        else:
            totals_trim_spec = [
                f
                for f in rebin.trim_spec_composer(
                    [self.ibins, self.jbins, self.kbins],
                    [xmin, ymin, zmin],
                    [xmax, ymax, zmax],
                )
            ]
            _, new_totals = rebin.shrink_nd(
                self.totals, iter(totals_trim_spec), assume_sorted=True
            )
            _, new_totals_err = rebin.shrink_nd(
                self.totals_err, iter(totals_trim_spec), assume_sorted=True
            )

        return FMesh(
            new_name,
            self.kind,
            gc.CartesianGeometrySpec(new_xbins, new_ybins, new_zbins),
            new_ebins,
            new_data,
            new_errors,
            new_totals,
            new_totals_err,
        )

    def rebin(
        self,
        new_x: np.ndarray,
        new_y: np.ndarray,
        new_z: np.ndarray,
        new_name=-1,
        extra_process_threshold: int = 1000000,
    ) -> "FMesh":
        """Extract data for a new spatial grid.

        Args:
            new_x:
                A new binning over X axis.
            new_y:
                A new binning over Y axis.
            new_z:
                A new binning over Z axis.
            new_name: optional
                A name for the rebinned mesh to be created.
            extra_process_threshold: optional
                At which size of data use multiple Python processes

        Returns:
            mesh:
                New FMesh object with the rebinned data.
        """
        assert not self.is_cylinder, "Not implemented for cylinder meshes"

        if self.data.size < extra_process_threshold:
            return self.rebin_single(new_x, new_y, new_z, new_name)

        # To avoid huge memory allocations, iterate over energy with external processes
        pool = ndp.Pool(processes=4)
        data_rebin_spec = [
            i
            for i in rebin.rebin_spec_composer(
                [self.ibins, self.jbins, self.kbins],
                [new_x, new_y, new_z],
                axes=[0, 1, 2],
            )
        ]

        def iter_over_e(data):
            for i in range(self.e.size - 1):
                yield data[i], data_rebin_spec, True

        new_data = np.stack(pool.map(_expand_args, iter_over_e(self.data)), axis=0)
        t = self.data * self.errors
        new_errors = np.stack(pool.map(_expand_args, iter_over_e(t)), axis=0)
        new_errors /= new_data
        if self.totals is None:
            new_totals = None
            new_totals_err = None
        else:
            new_totals = rebin.rebin_nd(
                self.totals, data_rebin_spec, assume_sorted=True
            )
            t = self.totals * self.totals_err
            new_totals_err = rebin.rebin_nd(t, data_rebin_spec, assume_sorted=True)
            new_totals_err /= new_totals

        return FMesh(
            new_name,
            self.kind,
            gc.CartesianGeometrySpec(new_x, new_y, new_z),
            self.e,
            new_data,
            new_errors,
            new_totals,
            new_totals_err,
        )

    def rebin_single(
        self,
        new_x: np.ndarray,
        new_y: np.ndarray,
        new_z: np.ndarray,
        new_name: int = -1,
    ) -> "FMesh":
        """Create FMesh object corresponding to this one by fluxes, but over new mesh.

        Ags:
            new_x:
                A new binning over X axis.
            new_y: ndarray
                A new binning over Y axis.
            new_z: ndarray
                A new binning over Z axis.
            new_name: int, optional
                name for the rebinned mesh to be created.

        Returns:
            mesh:
                New FMesh object with the rebinned data.
        """
        assert not self.is_cylinder, "Not implemented for cylinder meshes"

        data_rebin_spec = [
            i
            for i in rebin.rebin_spec_composer(
                [self.ibins, self.jbins, self.kbins],
                [new_x, new_y, new_z],
                axes=[1, 2, 3],
            )
        ]
        new_data = rebin.rebin_nd(self.data, iter(data_rebin_spec), assume_sorted=True)
        t = self.data * self.errors
        new_errors = rebin.rebin_nd(t, iter(data_rebin_spec), assume_sorted=True)
        new_errors /= new_data
        if self.totals is None:
            new_totals = None
            new_totals_err = None
        else:
            totals_rebin_spec = [
                i
                for i in rebin.rebin_spec_composer(
                    [self.ibins, self.jbins, self.kbins],
                    [new_x, new_y, new_z],
                    axes=[0, 1, 2],
                )
            ]
            new_totals = rebin.rebin_nd(
                self.totals, iter(totals_rebin_spec), assume_sorted=True
            )
            t = self.totals * self.totals_err
            new_totals_err = rebin.rebin_nd(
                t, iter(totals_rebin_spec), assume_sorted=True
            )
            new_totals_err /= new_totals

        return FMesh(
            new_name,
            self.kind,
            gc.CartesianGeometrySpec(new_x, new_y, new_z),
            self.e,
            new_data,
            new_errors,
            new_totals,
            new_totals_err,
        )

    def format_cylinder_origin_and_axis_label(self):
        if self.is_cylinder:
            return "\n  Cylinder origin at {0} {1} {2}, axis in {3} {4} {5} direction\n".format(
                self._geometry_spec.origin[0],
                self._geometry_spec.origin[1],
                self._geometry_spec.origin[2],
                self._geometry_spec.axs[0],
                self._geometry_spec.axs[1],
                self._geometry_spec.axs[2],
            )
        return ""


# noinspection PyTypeChecker,PyProtectedMember
def merge_tallies(
    name: int, kind: int, *tally_weight: Tuple[FMesh, float], comment: str = None
) -> FMesh:
    """Makes superposition of tallies with specific weights.

    Args:
        name:
            Name of new fmesh tally.
        kind:
            Type of new fmesh tally. It can be -1 (or any arbitrary integer).
        tally_weight:
            List of tally-weight pairs (tuples). tally is FMesh instance. weight
            is float.
        comment:
            A comment to assign to the new mesh tally

    Returns:
        result :
            The merged FMesh.
    """
    result_data = None
    errors = None
    geometry_spec = None
    ebins = None
    for t, w in tally_weight:  # type: FMesh, float
        if result_data is None:
            result_data = t.data * w
            errors = (t.errors * t.data * w) ** 2
            geometry_spec = t._geometry_spec
            ebins = t.e
        else:
            result_data += t.data * w
            errors += (t.errors * t.data * w) ** 2
            assert geometry_spec == t._geometry_spec
            assert np.array_equal(
                ebins.size, t.e.size
            )  # allow merging neutron and photon heating meshes
    nonzero_idx = np.logical_and(result_data > 0.0, errors > 0.0)
    result_error = np.zeros_like(result_data)
    result_error[nonzero_idx] = np.sqrt(errors[nonzero_idx]) / result_data[nonzero_idx]
    return FMesh(
        name,
        kind,
        geometry_spec,
        ebins,
        result_data,
        result_error,
        comment=comment,
    )


# def read_meshtally(file_):
#     """Reads fmesh tally from binary file_.
#
#     Parameters
#     ----------
#     file_ : file or str
#         File or filename from which tally should be loaded.
#
#     Returns
#     -------
#     mesh : FMesh
#         Fmesh tally instance.
#     """
#     data = np.load(file_)
#     ne = int(data[0])
#     nx = int(data[1])
#     ny = int(data[2])
#     nz = int(data[3])
#     name = int(data[4])
#     kind = int(data[5])
#     ebins = data[6 : ne + 6]
#     xbins = data[ne + 6 : ne + nx + 6]
#     ybins = data[ne + nx + 6 : ne + nx + ny + 6]
#     zbins = data[ne + nx + ny + 6 : ne + nx + ny + nz + 6]
#     n = (ne - 1) * (nx - 1) * (ny - 1) * (nz - 1)
#     sti = ne + nx + ny + nz + 6
#     data_f = data[sti : sti + n].reshape((ne - 1, nx - 1, ny - 1, nz - 1))
#     data_err = data[sti + n :].reshape((ne - 1, nx - 1, ny - 1, nz - 1))
#     return FMesh(name, kind, xbins, ybins, zbins, ebins, data_f, data_err)


def read_meshtal(stream: TextIO, select=None, mesh_file_info=None) -> List[FMesh]:
    """Reads fmesh tallies from a stream.

    Args:
        stream : The text stream to read.
        select  : predicate
            Selects the meshes actually to process
        mesh_file_info:
            object to collect information from m-file header

    Returns:
        tallies :
            The list of individual fmesh tally.
    """
    next(stream)  # TODO dvp check if we need to store problem time stamp
    next(stream)  # TODO dvp check if we need to store problem title
    line = next(stream)
    nps = int(float((line.strip().split("=")[1])))
    if mesh_file_info is not None:
        mesh_file_info.nps = nps
    return list(iter_meshtal(stream, select))


# noinspection PyTypeChecker
def iter_meshtal(
    fid: TextIO,
    name_select: Callable[[int], bool] = None,
    tally_select: Callable[[FMesh], bool] = None,
) -> Generator[FMesh, None, None]:
    """Iterates fmesh tallies from fid.

    Args:
        fid : A stream to read meshes from.
        name_select:  A function returning True, if tally name is acceptable, otherwise skips tally reading and parsing
        tally_select: A function returning True, if total tally content is acceptable

    Returns:
        iterator: An iterator over meshtal file with proper filtering over names or tallies content.
    """
    try:
        while True:
            # Skip first two comment lines ms version and model title
            # noinspection PyUnresolvedReferences
            name = int(_find_words_after(fid, "Mesh", "Tally", "Number")[0])
            if not name_select or name_select(name):
                # __LOG.debug("Reading mesh tally %s", name)
                comment = fid.readline().strip()
                if comment.startswith("This is a"):
                    kind = comment.split()[3]
                    comment = None
                else:
                    # noinspection PyUnresolvedReferences
                    kind = _find_words_after(fid, "This", "is", "a")[0]
                if comment:
                    comment = fix_mesh_comment(name, comment)
                kind = Kind[kind]

                # TODO dvp read "dose function modified" here

                _find_words_after(fid, "Tally", "bin", "boundaries:")

                line = next(fid).lstrip()
                if line.startswith("Cylinder"):
                    # retrieve cylinder origin and axis
                    part1, part2 = line.split(",")
                    origin = np.fromiter(part1.split()[3:6], dtype=float)
                    axis = np.fromiter(part2.split()[2:5], dtype=float)
                    ibins = np.array(
                        [
                            float(w)
                            for w in _find_words_after(
                                concatv([line], fid), "R", "direction:"
                            )
                        ]
                    )

                    jbins = np.array(
                        [float(w) for w in _find_words_after(fid, "Z", "direction:")]
                    )

                    kbins = np.array(
                        [
                            float(w)
                            for w in _find_words_after(
                                fid, "Theta", "direction", "(revolutions):"
                            )
                        ]
                    )

                    geometry_spec = gc.CylinderGeometrySpec(
                        ibins, jbins, kbins, origin=origin, axs=axis
                    )

                    ebins = np.array(
                        [
                            float(w)
                            for w in _find_words_after(
                                fid, "Energy", "bin", "boundaries:"
                            )
                        ]
                    )
                    with_ebins = check_ebins(
                        fid, ["Energy", "R", "Z", "Th", "Result", "Rel", "Error"]
                    )
                else:
                    xbins = np.array(
                        [
                            float(w)
                            for w in _find_words_after(
                                concatv([line], fid), "X", "direction:"
                            )
                        ]
                    )

                    ybins = np.array(
                        [float(w) for w in _find_words_after(fid, "Y", "direction:")]
                    )

                    zbins = np.array(
                        [float(w) for w in _find_words_after(fid, "Z", "direction:")]
                    )

                    geometry_spec = gc.CartesianGeometrySpec(xbins, ybins, zbins)

                    ebins = np.array(
                        [
                            float(w)
                            for w in _find_words_after(
                                fid, "Energy", "bin", "boundaries:"
                            )
                        ]
                    )
                    with_ebins = check_ebins(
                        fid, ["Energy", "X", "Y", "Z", "Result", "Rel", "Error"]
                    )

                spatial_bins_size = geometry_spec.bins_size
                bins_size = spatial_bins_size * (ebins.size - 1)

                def _iterate_bins(stream, n_, _with_ebins):
                    value_start, value_end = (41, 53) if _with_ebins else (32, 44)
                    for i in range(n_):
                        _line = next(stream)
                        _value = float(_line[value_start:value_end])
                        _error = float(_line[value_end:])
                        if _value < 0.0:
                            _value = _error = 0.0
                        yield _value
                        yield _error

                data_items = np.fromiter(
                    _iterate_bins(fid, bins_size, with_ebins), dtype=float
                )
                data_items = data_items.reshape(bins_size, 2)
                shape = (ebins.size - 1,) + geometry_spec.bins_shape
                data, error = data_items[:, 0].reshape(shape), data_items[:, 1].reshape(
                    shape
                )

                # reading totals for energy
                def _iterate_totals(stream, totals_number):
                    for i in range(totals_number):
                        _line = next(stream).split()
                        # TODO dvp: check for negative values in an MCNP meshtal file
                        assert "Total" == _line[0]
                        for w in _line[4:]:
                            yield float(w)

                if (
                    ebins.size > 2
                ):  # Totals are not output if there's only one bin in energy domain
                    totals_items = np.fromiter(
                        _iterate_totals(fid, spatial_bins_size), dtype=float
                    )
                    totals_items = totals_items.reshape(spatial_bins_size, 2)
                    shape = geometry_spec.bins_shape
                    totals = totals_items[:, 0].reshape(shape)
                    totals_err = totals_items[:, 1].reshape(shape)
                else:
                    totals = None
                    totals_err = None
                res = FMesh(
                    name,
                    kind,
                    geometry_spec,
                    ebins,
                    data,
                    error,
                    totals,
                    totals_err,
                    comment=comment,
                )
                if not tally_select or tally_select(res):
                    yield res
                else:
                    __LOG.debug("Skipping mesh tally %s", name)
    except EOFError:
        pass


def check_ebins(fid: Iterable[str], keys: List[str]) -> bool:
    """Check if energy bins present in a mesh tally output values.

    If next nonempty line starts with a word keys[0] (i.e. "Energy"), then the energy bins present.
    Also check that the remaining keys correspond to the nonempty line.

    Args:
        fid: text rows to scan, including prepending empty rows
        keys: sequence of words to check

    Returns:
        bool: True if energy bins are present, False otherwise.

    Raises:
        ValueError: if keys don't correspond to the nonempty line.

    """
    title_line = _next_not_empty_line(fid)
    if title_line is None:
        raise ValueError(f"Cannot find titles {keys[1:]}")
    if title_line[0] == keys[0]:
        assert keys[1:] == title_line[1:]
        with_ebins = True
    else:
        if keys[1:] != title_line:
            raise ValueError(f"Unexpected values title {title_line}")
        with_ebins = False
    return with_ebins


def _next_not_empty_line(f: Iterable[str]) -> Optional[List[str]]:
    for line in f:
        words = line.split()
        if 0 < len(words):
            return words
    return None


def _find_words_after(f, *keywords: str) -> list[str]:
    """Searches for words that follow keywords.

    The line from file f is read. Then it is split into words (by spaces).
    If its first words are the same as keywords, then remaining words (up to
    newline character) are returned. Otherwise, new line is read.

    Args:
        f: File in which words are searched.
        keywords:  List of keywords after which right words are. The order is important.

    Returns:
        The list of words that follow keywords.
    """
    for line in f:
        words: list[str] = line.split()
        i = 0
        for w, kw in zip(words, keywords):
            if w != kw:
                break
            i += 1
        if i >= len(keywords):
            return words[i:]
    raise EOFError


def m_2_npz(
    stream: TextIO,
    prefix: Path,
    *,
    name_select=lambda _: True,
    tally_select=lambda _: True,
    suffix: str = "",
    mesh_file_info=None,
    check_existing_file_strategy=raise_error_when_file_exists_strategy,
):
    """Splits the tallies from the mesh file into separate npz files.

    Args:
        stream: File with MCNP mesh tallies
        prefix: Prefix for separate mesh files names
        name_select: function(int)->bool
            Filter fmesh by names
        tally_select: function(FMesh)->bool
            Filter fmesh by content.
        suffix: srt
            Prefix for separate mesh files names
        mesh_file_info: structure to store meshtal file header info: nps.
        check_existing_file_strategy: what to do if an output file already exists

    Returns:
        total:  Total number of files created
    """
    total = 0
    next(stream)  # TODO dvp check if we need to store problem time stamp
    next(stream)  # TODO dvp check if we need to store problem title
    line = next(stream)
    nps = int(float((line.strip().split("=")[1])))
    if mesh_file_info is not None:
        mesh_file_info.nps = nps

    for t in iter_meshtal(stream, name_select=name_select, tally_select=tally_select):
        t.save_2_npz(prefix / (str(t.name) + suffix), check_existing_file_strategy)
        total += 1

    return total


def fix_mesh_comment(mesh_no: int, comment: str) -> str:
    str_mesh_no = f"{mesh_no}"
    chars_to_remove = len(str_mesh_no) - 3
    if chars_to_remove > 0:
        comment = comment[chars_to_remove:]
    return comment.strip()


def meshes_to_vtk(
    *meshes: FMesh,
    out_dir: Path = None,
    get_mesh_description_strategy: Callable[[FMesh], str],
) -> None:
    if out_dir:
        out_dir.mkdir(parents=True, exist_ok=True)
    for mesh in meshes:
        particle = mesh.kind.short
        function = get_mesh_description_strategy(mesh)
        data_name = f"{particle}-{function}"
        file_name = f"{data_name}-{mesh.name}"
        if out_dir:
            file_name = str(out_dir / file_name)
        mesh.save2vtk(file_name, data_name)



from typing import Any, Generator, List, NamedTuple, Optional, TextIO, Tuple, Union

import sys

from dataclasses import dataclass
from enum import IntEnum

import mckit_meshes.mesh.geometry_spec as gs
import mckit_meshes.utils as ut
import numpy as np

GeometrySpec = Union[gs.CartesianGeometrySpec, gs.CylinderGeometrySpec]
Point = np.ndarray


def ensure_float_arrays(*arrays: Any) -> Generator[np.ndarray, None, None]:
    yield from map(lambda x: np.asarray(x, dtype=float), arrays)


class Particles(IntEnum):
    neutron = 0
    photon = 1
    n = 0
    p = 1


# noinspection GrazieInspection
class WgtMesh:
    """
    Class to work with MCNP weight window files.
    """

    def __init__(
        self,
        geometry_spec: GeometrySpec,
        energies,
        weights,
    ):
        self._energies: List[np.ndarray] = list(ensure_float_arrays(*energies))
        self._geometry_spec = geometry_spec
        self._weights: List[np.ndarray] = list(ensure_float_arrays(*weights))
        self.validate()

    def print_mcnp_generator_spec(self, io=None, ref="600 0 50", columns: int = 6):
        if io is None:
            io = sys.stdout
        print(f"mesh   ref={ref}", file=io)
        self._geometry_spec.print(io, columns=columns)
        print("wwge:n", end=" ", file=io)
        second_indent = " " * 15
        ut.print_n(
            gs.format_floats(self.energies[0][1:]),
            io=io,
            indent=second_indent,
            columns=columns,
        )
        if len(self.energies) > 1:
            print("wwge:p", end=" ", file=io)
            ut.print_n(
                gs.format_floats(self.energies[1][1:]),
                io=io,
                indent=second_indent,
                columns=columns,
            )

    def print_meshtal_spec(
        self,
        io: TextIO = None,
        tally_n_number: int = 14,
        tally_p_number: int = 24,
        columns: int = 6,
    ) -> None:
        if io is None:
            io = sys.stdout
        print(f"fc{tally_n_number}    === WW generation mesh for neutrons", file=io)
        print(f"fmesh{tally_n_number}:n", file=io)
        self._geometry_spec.print(io, columns=columns)
        indent = " " * 8
        print(indent, "emesh=", sep="", end="", file=io)
        second_indent = indent + " " * 6
        ut.print_n(
            gs.format_floats(self.energies[0][1:]),
            io=io,
            indent=second_indent,
            columns=columns,
        )
        if len(self.energies) > 1:
            print(f"fc{tally_p_number}    === WW generation mesh for photons", file=io)
            print(f"fmesh{tally_p_number}:p", file=io)
            self._geometry_spec.print(io, columns=columns)
            print(indent, "emesh=", sep="", end="", file=io)
            # TODO dvp: try to use do_print_bins here
            ut.print_n(
                gs.format_floats(self.energies[1][1:]),
                io=io,
                indent=second_indent,
                columns=columns,
            )

    def validate(self):
        if len(self.weights) != len(self.energies):
            raise ValueError(
                f"Number of energy bins {len(self.energies)} is not equal to number of weight parts {len(self.weights)}"
            )
        for part, ebins in enumerate(self.energies):
            expected_shape = tuple(
                [
                    ebins.size - 1,
                    self.ibins.size - 1,
                    self.jbins.size - 1,
                    self.kbins.size - 1,
                ]
            )
            if self.weights[part].shape != expected_shape:
                raise ValueError(
                    f"Incompatible number of ebins, voxels and weights: {self.weights[part].shape} != {expected_shape}"
                )

    @property
    def energies(self):
        return self._energies

    @property
    def origin(self):
        return self._geometry_spec.origin

    @property
    def ibins(self):
        return self._geometry_spec.ibins

    @property
    def jbins(self):
        return self._geometry_spec.jbins

    @property
    def kbins(self):
        return self._geometry_spec.kbins

    @property
    def count_voxels(self):
        return (self.ibins.size - 1) * (self.jbins.size - 1) * (self.kbins.size - 1)

    @property
    def weights(self) -> List[np.ndarray]:
        return self._weights

    @property
    def neutron_weights(self) -> np.ndarray:
        return self._weights[0]

    @property
    def photon_weights(self) -> np.ndarray:
        assert len(self._weights) == 2, "Photon weights are not defined in the mesh"
        return self._weights[1]

    @property
    def is_cylinder(self) -> bool:
        return self._geometry_spec.cylinder

    @property
    def axs(self) -> Optional[np.ndarray]:
        return self._geometry_spec.axs

    @property
    def vec(self) -> Optional[np.ndarray]:
        return self._geometry_spec.vec

    @property
    def count_parts(self) -> int:
        return len(self.weights)

    def part(self, particle: Particles) -> Tuple[np.ndarray, np.ndarray]:
        return self.energies[particle], self.weights[particle]

    def __hash__(self):
        return hash(
            (
                self._geometry_spec.__hash__(),
                self.energies,
                self.weights,
            )
        )

    def __eq__(self, other):
        if not self.bins_are_equal(other):
            return False
        for p in range(len(self.weights)):
            if not np.array_equal(self.weights[p], other.weights[p]):
                return False
        return True

    def bins_are_equal(self, other: "WgtMesh") -> bool:
        if not isinstance(other, WgtMesh):
            raise RuntimeError(
                "Invalid class of object to compare: {}", other.__class__
            )
        if self._geometry_spec == other._geometry_spec:
            le = len(self.energies)
            if le == len(other.energies):
                for i in range(le):
                    if not np.array_equal(self.energies[i], other.energies[i]):
                        return False
                return True
        return False

    def __add__(self, other) -> "WgtMesh":
        assert self.bins_are_equal(other)
        weights = [a + b for a, b in zip(self.weights, other.weights)]
        res = WgtMesh(
            self._geometry_spec,
            self.energies,
            weights,
        )
        return res

    def __sub__(self, other) -> "WgtMesh":
        assert self.bins_are_equal(other)
        weights = [a - b for a, b in zip(self.weights, other.weights)]
        res = WgtMesh(
            self._geometry_spec,
            self.energies,
            weights,
        )
        return res

    def __mul__(self, coeff: float) -> "WgtMesh":
        weights = [w * coeff for w in self.weights]
        res = WgtMesh(
            self._geometry_spec,
            self.energies,
            weights,
        )
        return res

    def __rmul__(self, coeff: float) -> "WgtMesh":
        return self.__mul__(coeff)

    # def __repr__(self):
    #     return f"WgtMesh({tuple(self.origin)}, {self.energies})"

    # noinspection SpellCheckingInspection
    def write(self, stream: TextIO) -> None:
        """Writes the mesh to stream.

        See WWINP format, MCNP User Manual, Appendix J, Table J.1

        Args;
            stream: a stream to write to

        """
        data = []
        _if, _iv, _ni = 1, 1, len(self.energies)
        if self.is_cylinder:
            _nr = 16
        else:
            _nr = 10
        data += produce_strings([_if, _iv, _ni, _nr], "{0:10d}")
        # remove the first "\n"
        data = data[1:]

        _ne = [x.size - 1 for x in self._energies]
        data += produce_strings(_ne, "{0:10d}")
        _nfx = self.ibins.size - 1
        _nfy = self.jbins.size - 1
        _nfz = self.kbins.size - 1
        _x0, _y0, _z0 = self.origin
        _nfmx, _xm = gs.compute_intervals_and_coarse_bins(self.ibins)
        _ncx = len(_nfmx)
        _nfmy, _ym = gs.compute_intervals_and_coarse_bins(self.jbins)
        _ncy = len(_nfmy)
        _nfmz, _zm = gs.compute_intervals_and_coarse_bins(self.kbins)
        _ncz = len(_nfmz)
        _nwg = 1
        _data = [_nfx, _nfy, _nfz, _x0, _y0, _z0, _ncx, _ncy, _ncz]
        if self.is_cylinder:
            if self.axs is None:
                raise ValueError("axs is not specified in cylinder mesh")
            _xmax, _ymax, _zmax = self.axs
            if self.vec is None:
                raise ValueError("vec is not specified in cylinder mesh")
            _xr, _yr, _zr = self.vec
            _data += [_xmax, _ymax, _zmax, _xr, _yr, _zr]
            _nwg = 2
        _data += [_nwg]
        data += produce_strings(_data, "{0:#13.5g}")
        # Block 2
        _nc = [_ncx, _ncy, _ncz]
        _nfm = [_nfmx, _nfmy, _nfmz]
        _r = [_xm, _ym, _zm]
        for i in range(3):
            data1 = [_r[i][0]]
            for j in range(_nc[i]):
                data1 += [_nfm[i][j], _r[i][j + 1], 1]
            data += produce_strings(data1, "{0:#13.5g}")
        for p in range(_ni):
            data += produce_strings(
                self.energies[p][1:], "{0:#13.5g}"
            )  # omit the first zero
            data1 = []
            # TODO dvp: order of cycling and dimensions are not efficient:
            #           the index changing faster should be the most inner in cycle.
            for e in range(_ne[p]):
                for k in range(_nfz):
                    for j in range(_nfy):
                        for i in range(_nfx):
                            data1.append(self._weights[p][e, i, j, k])
            data += produce_strings(data1, "{0:#13.5g}")
        stream.write("".join(data))

    @dataclass
    class _Reader:
        data: List[str]
        index: int = 0

        def get(self, items: int) -> List[str]:
            i = self.index
            self.index += items
            return self.data[i : self.index]

        def get_floats(self, items: int) -> Generator[float, None, None]:
            return map(float, self.get(items))

        def get_ints(self, items: int) -> Generator[int, None, None]:
            return map(int, self.get(items))

        def get_ints_written_as_floats(self, items: int) -> Generator[int, None, None]:
            return map(int, self.get_floats(items))

        def skip(self, items: int = 1) -> None:
            self.index += items

    # noinspection SpellCheckingInspection
    @classmethod
    def read(cls, f: TextIO) -> "WgtMesh":
        """Read an MCNP weights file.

        See format description at MCNP User Manual, Version 5 (p.489 or Appendix J, p. J-1)

        Args:
            f: Input file in WWINP format

        Returns:
            WgtMesh: loaded mesh.

        """
        _if, _iv, number_of_particles, number_of_parameters = (
            int(s) for s in f.readline().split()[:4]
        )

        reader = WgtMesh._Reader(f.read().split())
        sizes_of_energy_bins = tuple(reader.get_ints(number_of_particles))

        # cells along axes
        _nfx, _nfy, _nfz = reader.get_ints_written_as_floats(3)

        # origin
        _x0, _y0, _z0 = reader.get_floats(3)

        # coarse bins along axes
        _ncx, _ncy, _ncz = reader.get_ints_written_as_floats(3)
        if number_of_parameters == 16:
            _xmax, _ymax, _zmax, _xr, _yr, _zr = reader.get_floats(6)
            axs = np.array([_xmax, _ymax, _zmax], dtype=float)
            vec = np.array([_xr, _yr, _zr], dtype=float)
        else:
            axs = None
            vec = None

        # skip NWG
        reader.skip()

        delta = 3 * _ncx + 1
        _x = parse_coordinates(reader.get(delta))

        delta = 3 * _ncy + 1
        _y = parse_coordinates(reader.get(delta))

        delta = 3 * _ncz + 1
        _z = parse_coordinates(reader.get(delta))

        _e = []
        _w = []

        for p in range(number_of_particles):
            nep = sizes_of_energy_bins[p]
            if 0 < nep:
                ebins = np.fromiter(reader.get_floats(nep), dtype=float)
                ebins = np.insert(ebins, 0, 0.0)
                _e.append(ebins)
                _wp = np.zeros((nep, _nfx, _nfy, _nfz), dtype=float)
                _wp_data = np.fromiter(reader.get_floats(_wp.size), dtype=float)
                for e in range(nep):
                    for k in range(_nfz):
                        for j in range(_nfy):
                            for i in range(_nfx):
                                cell_index = i + _nfx * (j + _nfy * (k + _nfz * e))
                                _wp[e, i, j, k] = _wp_data[cell_index]
                _w.append(_wp)
                assert np.all(
                    np.transpose(
                        _wp_data.reshape((nep, _nfz, _nfy, _nfx)), (0, 3, 2, 1)
                    )
                    == _wp
                )
        geometry_spec = make_geometry_spec(
            [_x0, _y0, _z0], _x, _y, _z, axs=axs, vec=vec
        )
        return cls(geometry_spec, _e, _w)

    def get_mean_square_distance_weights(self, point):
        w = self._geometry_spec.get_mean_square_distance_weights(point)
        _w = []
        for _e in self.energies:
            le = len(_e)
            t = w.reshape((1,) + self._geometry_spec.bins_shape)
            t = np.repeat(t, le, axis=0)
            _w.append(t)
        return WgtMesh(
            self._geometry_spec,
            self.energies,
            _w,
        )

    class MergeSpec(NamedTuple):
        wm: "WgtMesh"
        nps: int

    @classmethod
    def merge(cls, *merge_specs: Union[MergeSpec, Tuple["WgtMesh", int]]) -> MergeSpec:
        r"""Combine weight meshes produced from different runs with weighting factor.

        Note:

        Importance of a mesh voxel `i` is $1/w_i$ and is proportional to average portion $p_i$ of
        passing particle weight W to a tally for which the weight mesh is computed.
        To obtain combined weight on merging two meshes, we will combine the probabilities using weighting factors and
        use reciprocal of a result as a resulting weight of mesh voxel.
        The weighting factors are usually NPS (Number particles sampled) from a run on which a mesh was produced.
        The combined probability in resulting voxel `i` is:

        .. math::

            w_ij - weight in voxel i of mesh j
            n_j -  nps - weighting factor on combining of mesh j

            p_ij = 1/w_ij - probability for voxel i of mesh j

            p_i = \frac{ \sum_j{n_j*p_ij} { \sum_j{n_j} }

        So, the resulting voxel `i` weight level is:

        .. math::
            w_i = \frac{1} {p_i}


        Args:
            merge_specs: iterable of pairs (WgtMesh, nps), where `nps` is weighting factor

        Returns:
            MergeSpec: merged weights and total nps (or sum of weighting factors)



        """
        first = merge_specs[0]

        if not isinstance(first, WgtMesh.MergeSpec):
            first = WgtMesh.MergeSpec(*first)  # convert tuple to MergeSpec

        if len(merge_specs) > 1:
            second = WgtMesh.merge(*merge_specs[1:])
            merged_weights = []
            assert first.wm.bins_are_equal(second.wm)
            for i, weights in enumerate(first.wm.weights):
                nps_first, probabilities_first = prepare_probabilities_and_nps(
                    first.nps, weights
                )
                nps_second, probabilities_second = prepare_probabilities_and_nps(
                    second.nps, second.wm.weights[i]
                )
                nps = np.array(nps_first + nps_second, dtype=float)
                combined_probabilities = (
                    nps_first * probabilities_first + nps_second * probabilities_second
                ) * reciprocal(nps)
                merged_weights.append(reciprocal(combined_probabilities))
            wm = first.wm
            return WgtMesh.MergeSpec(
                cls(
                    wm._geometry_spec,
                    wm.energies,
                    merged_weights,
                ),
                first.nps + second.nps,
            )
        else:
            return first

    def reciprocal(self) -> "WgtMesh":
        """
        Invert weights values.

        To be used for anti-forward method of weight generation.

        Returns
        -------
        out:
            Reciprocal of this weights
        """
        return WgtMesh(
            self._geometry_spec, self.energies, list(map(reciprocal, self.weights))
        )

    def normalize(
        self, normalization_point: Point, normalized_value: float = 1.0, energy_bin=-1
    ) -> "WgtMesh":
        """
        Scale weights to have value `value` at `normalisation_point`.

        All other voxels are scaled proportionally.

        Args:
            normalization_point:  Coordinates of point where the weights should equal `value`.
            normalized_value: The value which should be at `normalization_point`
            energy_bin: index of energy bin at which set normalized value, default - the last one.

        Returns:
            New normalized weights.
        """
        gs = self._geometry_spec
        x, y, z = normalization_point
        ix, iy, iz = gs.select_indexes(i_values=x, j_values=y, k_values=z)
        new_weights = []
        value_at_normalisation_point = self.weights[0][energy_bin, ix, iy, iz]
        """The value at last energy bin about 20 MeV at neutron weights."""

        factor = normalized_value / value_at_normalisation_point
        """Scale all other weights by this value."""

        for i, w in enumerate(self.weights):
            # TODO dvp: revise for multiple energy bins, may be add scaling values for each energy bin and particle
            value_at_normalisation_point = w[
                energy_bin, ix, iy, iz
            ]  # the last energy bin about 20 MeV.
            new_weights.append(w * factor)

        return WgtMesh(gs, self.energies, new_weights)

    def invert(
        self, normalization_point: Point, normalized_value: float = 1.0
    ) -> "WgtMesh":
        """Get reciprocal of self weights and normalize to 1 at given point.

        Important:
            A caller specifies normalization_point in local coordinates. See :class:`GeometrySpec.local_coordinates`.

        Args:
            normalization_point: Point at which output weights should be 1
            normalized_value: value which should be set at `normalization_point`.

        Returns:
            WgtMesh: Normalized reciprocal of self weights.

        """
        return self.reciprocal().normalize(normalization_point, normalized_value)

    @property
    def geometry_spec(self):
        return self._geometry_spec

    def drop_lower_energies(self, min_energy: float, part: int = 0) -> "WgtMesh":
        if len(self.energies) <= part:
            raise ValueError(f"invalid value for weights object part: {part}")
        energies = self.energies[part]
        energies_to_retain = min_energy <= energies
        energies_to_retain[0] = True
        if np.all(energies_to_retain):
            return self
        else:
            new_energies = []
            new_weights = []
            for i in range(len(self.energies)):
                if i == part:
                    new_energies.append(self.energies[i][energies_to_retain])
                    new_weights.append(self.weights[i][energies_to_retain[1:], :, :, :])
                else:
                    new_energies.append(self.energies[i])
                    new_weights.append(self.weights[i])
            gs = self._geometry_spec
            return WgtMesh(gs, new_energies, new_weights)


def reciprocal(a: np.ndarray, zero_index: np.ndarray = None) -> np.ndarray:
    if a.dtype != float:
        a = np.array(a, dtype=float)
    if zero_index is None:
        zero_index = a == 0.0
    else:
        assert np.array_equal(zero_index, a == 0.0)
    result: np.ndarray = np.reciprocal(a, where=np.logical_not(zero_index))
    #  fix bug in numpy reciprocal: it doesn't pass zero values, the bug doesn't show up on debugging
    result[zero_index] = 0.0
    return result


def prepare_probabilities_and_nps(
    _nps: int, _weights: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes intermediate data for merging procedure.

    The probabilities are reciprocals to weights.
    Zero weights mean zero probabilities and don't affect the merged result.

    Args:
        _nps:
            weighting multiplier
        _weights:
            weights to convert to probabilities
    Returns:
        normalization factors and probabilities
    """
    nps_array = np.full_like(_weights, _nps, dtype=int)
    zero_index = _weights == 0.0
    probabilities = reciprocal(_weights, zero_index=zero_index)
    nps_array[zero_index] = 0  # voxels with zero weight don't affect weighted sum
    return nps_array, probabilities


def produce_strings(stream, format_spec):
    data = []
    for i in range(len(stream)):
        if i % 6 == 0:
            data.append("\n")
        data.append(format_spec.format(stream[i]))
    return data


def parse_coordinates(inp: List[str]) -> np.ndarray:
    def iter_over_coarse_mesh():
        is_first = True
        i = 0
        length = len(inp)
        while i < length:
            coordinate = float(inp[i])
            if is_first:
                i += 1
                is_first = False
            else:
                i += 2
            if length <= i:
                yield coordinate, 1
                break
            else:
                fine_bins = int(float(inp[i]))
                i += 1
            yield coordinate, fine_bins

    def iter_over_fine_mesh(_iter_over_coarse_mesh):
        prev_coordinate = None
        prev_fine_bins = None
        for coordinate, fine_bins in _iter_over_coarse_mesh:
            if prev_fine_bins == 1:
                yield prev_coordinate
            elif prev_coordinate is not None:
                res = np.linspace(
                    prev_coordinate,
                    coordinate,
                    prev_fine_bins + 1,
                    endpoint=True,
                    dtype=float,
                )
                for coord in res[:-1]:
                    yield coord
            prev_coordinate = coordinate
            prev_fine_bins = fine_bins
        yield prev_coordinate

    return np.fromiter(iter_over_fine_mesh(iter_over_coarse_mesh()), dtype=float)


def make_geometry_spec(origin, ibins, jbins, kbins, axs=None, vec=None) -> GeometrySpec:
    origin, ibins, jbins, kbins = map(gs.as_float_array, [origin, ibins, jbins, kbins])
    if axs is None:
        geometry_spec = gs.CartesianGeometrySpec(ibins, jbins, kbins, origin=origin)
    else:
        axs, vec = map(gs.as_float_array, [axs, vec])
        geometry_spec = gs.CylinderGeometrySpec(
            ibins, jbins, kbins, origin=origin, axs=axs, vec=vec
        )
    return geometry_spec
